# -*- coding: utf-8 -*-
"""
    weppy.tools.auth.handlers
    -------------------------

    Provides handlers for the authorization system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from datetime import timedelta
from ...dal import Field
from ...forms import Form
from ...globals import request, session
from ...handlers import Handler
from ...helpers import flash
from ...http import redirect
from ...language import T


class AuthLoginHandler(object):
    store_password = False
    create_user_onlogin = True
    next = None

    def __init__(self, auth, env=None):
        self.auth = auth
        self.env = env
        self.user = None

    def onsuccess(self):
        pass

    def get_user(self):
        return None


class DefaultLoginHandler(AuthLoginHandler):
    store_password = True
    create_user_onlogin = False

    def __init__(self, auth, env):
        AuthLoginHandler.__init__(self, auth, env)
        self.userfield = self.auth.settings.login_userfield
        # TODO: labels
        if self.userfield == 'email':
            self.loginfield = Field(
                validation={'is': 'email', 'presence': True}
            )
        else:
            v = {'presence': True}
            if self.auth.settings.username_case_sensitive:
                v['lower'] = True
            self.loginfield = Field(validation=v)
        passfield_valid = self.auth.settings.models.user.password._requires
        self.passfield = Field('password', validation=passfield_valid)
        self.rememberfield = Field(
            'bool', default=True, label=T('Remember me')
        )

    def get_user(self):
        return self.user

    def onaccept(self, form):
        userfield = self.userfield
        #passfield = self.passfield
        entered_username = form.params[userfield]
        #if multi_login and '@' in entered_username:
        #   # if '@' in username check for email, not username
        #   user = self.table_user(email = entered_username)
        #else:
        user = self.auth.table_user(**{userfield: entered_username})
        if user:
            # user in db, check if registration pending or disabled
            temp_user = user
            if temp_user.registration_key == 'pending':
                flash(self.auth.messages.registration_pending)
                return
            elif temp_user.registration_key in ('disabled', 'blocked'):
                flash(self.auth.messages.login_disabled)
                return
            elif temp_user.registration_key is not None and \
                    temp_user.registration_key.strip():
                flash(self.auth.messages.registration_verifying)
                return
            #: verify password
            if form.params.get('password', '') == temp_user.password:
                # success
                self.user = temp_user
        if not self.user:
            self.onfail()

    def onsuccess(self):
        flash(self.auth.messages.logged_in)

    def onfail(self):
        self.auth.log_event(self.auth.messages['login_failed_log'],
                            request.body_params)
        flash(self.auth.messages.invalid_login)
        redirect(self.auth.url('login', request.query_params))

    def login_form(self):
        userfield = self.userfield
        form_fields = {
            userfield: self.loginfield,
            'password': self.passfield
        }
        if self.auth.settings.remember_me_form:
            form_fields['remember'] = self.rememberfield
        form = Form(
            form_fields,
            hidden=dict(_next=self.auth.settings.login_next),
            submit=self.auth.messages.login_button
        )
        '''
        captcha = settings.login_captcha or \
            (settings.login_captcha != False and settings.captcha)
        if captcha:
            addrow(form, captcha.label, captcha, captcha.comment,
                   settings.formstyle, 'captcha__row')
        '''
        if form.accepted:
            self.onaccept(form)
            #: rebuild the form
            if not self.user:
                return self.login_form()
        return form


class AuthManager(Handler):
    def __init__(self, auth):
        #: the Auth() instance
        self.auth = auth
        user_table = self.auth.settings.models.user.table
        self._virtuals = [f.name for f in user_table._virtual_fields]
        self._virtualmethods = [m.name for m in user_table._virtual_methods]

    def _load_virtuals(self):
        self.auth.settings.models.user._inject_virtuals_on_row(self.auth.user)

    def _unload_virtuals(self):
        for namelist in [self._virtuals, self._virtualmethods]:
            for name in namelist:
                try:
                    del self.auth.user[name]
                except:
                    pass

    def on_start(self):
        # check auth session is valid
        authsess = self.auth._auth
        if authsess:
            #: check session has needed data
            if not authsess.last_visit or not authsess.last_dbcheck:
                del session.auth
                return
            #: is session expired?
            if (authsess.last_visit +
               timedelta(seconds=authsess.expiration) < request.now):
                del session.auth
            #: does session need re-sync with db?
            elif authsess.last_dbcheck + timedelta(seconds=360) < request.now:
                if self.auth.user:
                    #: is user still valid?
                    dbrow = self.auth.table_user(id=self.auth.user.id)
                    if dbrow and not dbrow.registration_key:
                        self.auth.login_user(
                            self.auth.table_user(id=self.auth.user.id),
                            authsess.remember)
                    else:
                        del session.auth
            else:
                #: set last_visit if make sense
                if ((request.now - authsess.last_visit).seconds >
                   (authsess.expiration / 10)):
                    authsess.last_visit = request.now
        #: load virtuals from Auth Model
        if self.auth.user:
            self._load_virtuals()

    def on_end(self):
        # set correct session expiration if requested by user
        if self.auth._auth and self.auth._auth.remember:
            session._expires_after(self.auth._auth.expiration)
        # remove virtual fields for serialization
        if self.auth.user:
            self._unload_virtuals()



# -*- coding: utf-8 -*-
"""
    weppy.debug
    -----------

    Provides debugging utilities.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Quite a lot of magic comes from Flask (http://flask.pocoo.org)
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

import traceback
import inspect
import os
import sys
from types import TracebackType, CodeType
from ._compat import PY2, reraise, iteritems
from .templating.helpers import TemplateError
from .utils import cachedprop


# on pypy we can take advantage of transparent proxies
try:
    from __pypy__ import tproxy
except ImportError:
    tproxy = None


# how does the raise helper look like?
try:
    exec("raise TypeError, 'foo'")
except SyntaxError:
    raise_helper = 'raise __weppy_exception__[1]'
except TypeError:
    raise_helper = 'raise __weppy_exception__[0], __weppy_exception__[1]'


class TracebackFrameProxy(object):
    """Proxies a traceback frame."""

    def __init__(self, tb):
        self.tb = tb
        self._tb_next = None

    @property
    def tb_next(self):
        return self._tb_next

    def set_next(self, next):
        if tb_set_next is not None:
            try:
                tb_set_next(self.tb, next and next.tb or None)
            except Exception:
                # this function can fail due to all the hackery it does
                # on various python implementations.  We just catch errors
                # down and ignore them if necessary.
                pass
        self._tb_next = next

    #@property
    #def is_weppy_frame(self):
    #    return '__weppy_template__' in self.tb.tb_frame.f_globals

    def __getattr__(self, name):
        return getattr(self.tb, name)


class ProcessedTraceback(object):
    """Holds a weppy preprocessed traceback for printing or reraising."""

    def __init__(self, exc_type, exc_value, frames):
        assert frames, 'no frames for this traceback?'
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.frames = frames

        # newly concatenate the frames (which are proxies)
        prev_tb = None
        for tb in self.frames:
            if prev_tb is not None:
                prev_tb.set_next(tb)
            prev_tb = tb
        prev_tb.set_next(None)

    def render_as_text(self, limit=None):
        """Return a string with the traceback."""
        lines = traceback.format_exception(self.exc_type, self.exc_value,
                                           self.frames[0], limit=limit)
        return ''.join(lines).rstrip()

    @property
    def exc_info(self):
        """Exception info tuple with a proxy around the frame objects."""
        return self.exc_type, self.exc_value, self.frames[0]

    @property
    def standard_exc_info(self):
        """Standard python exc_info for re-raising"""
        tb = self.frames[0]
        # the frame will be an actual traceback (or transparent proxy) if
        # we are on pypy or a python implementation with support for tproxy
        if type(tb) is not TracebackType:
            tb = tb.tb
        return self.exc_type, self.exc_value, tb


class Traceback(object):
    """Wraps a traceback."""

    def __init__(self, app, exc_type, exc_value, tb):
        self.app = app
        self.exc_type = exc_type
        self.exc_value = exc_value
        if not isinstance(exc_type, str):
            exception_type = exc_type.__name__
            if exc_type.__module__ not in ('__builtin__', 'exceptions'):
                exception_type = exc_type.__module__ + '.' + exception_type
        else:
            exception_type = exc_type
        self.exception_type = exception_type

        self.frames = []
        while tb:
            self.frames.append(Frame(self.app, exc_type, exc_value, tb))
            tb = tb.tb_next

    @property
    def exception(self):
        """String representation of the exception."""
        buf = traceback.format_exception_only(self.exc_type, self.exc_value)
        rv = ''.join(buf).strip()
        return rv.decode('utf-8', 'replace') if PY2 else rv

    def generate_plaintext_traceback(self):
        """Like the plaintext attribute but returns a generator"""
        yield u'Traceback (most recent call last):'
        for frame in self.frames:
            yield u'  File "%s", line %s, in %s' % (
                frame.filename,
                frame.lineno,
                frame.function_name
            )
            yield u'    ' + frame.current_line.strip()
        yield self.exception

    def generate_plain_tb_app(self):
        yield u'Traceback (most recent call last):'
        for frame in self.frames:
            if not frame.is_in_fw:
                yield u'  File "%s", line %s, in %s' % (
                    frame.filename,
                    frame.lineno,
                    frame.function_name
                )
                yield u'    ' + frame.current_line.strip()
        yield self.exception

    @property
    def full_tb(self):
        return u'\n'.join(self.generate_plaintext_traceback())

    @property
    def app_tb(self):
        return u'\n'.join(self.generate_plain_tb_app())


class Frame(object):
    """A single frame in a traceback."""

    def __init__(self, app, exc_type, exc_value, tb):
        self.app = app
        self.lineno = tb.tb_lineno
        self.function_name = tb.tb_frame.f_code.co_name
        self.locals = tb.tb_frame.f_locals
        self.globals = tb.tb_frame.f_globals

        fn = inspect.getsourcefile(tb) or inspect.getfile(tb)
        if fn[-4:] in ('.pyo', '.pyc'):
            fn = fn[:-1]
        # if it's a file on the file system resolve the real filename.
        if os.path.isfile(fn):
            fn = os.path.realpath(fn)
        self.filename = fn
        self.module = self.globals.get('__name__')
        #self.loader = self.globals.get('__loader__')
        self.code = tb.tb_frame.f_code

    @property
    def is_in_fw(self):
        fw_path = os.path.dirname(__file__)
        return self.filename[0:len(fw_path)] == fw_path

    @property
    def is_in_app(self):
        return self.filename[0:len(self.app.root_path)] == self.app.root_path

    @property
    def rendered_filename(self):
        if self.is_in_app:
            return self.filename[len(self.app.root_path) + 1:]
        if self.is_in_fw:
            return "weppy." + self.filename[len(os.path.dirname(__file__)) + 1:]\
                .replace("/", ".").split(".py")[0]
        return self.filename

    @cachedprop
    def sourcelines(self):
        try:
            with open(self.filename, 'rb') as file:
                source = file.read().decode('utf8')
        except IOError:
            source = '<unavailable>'
        return source.splitlines()

    @property
    def sourceblock(self):
        lmax = self.lineno + 4
        return u'\n'.join(self.sourcelines[self.first_line_no - 1:lmax])

    @property
    def first_line_no(self):
        l = self.lineno > 5 and (self.lineno - 5) or 1
        if l > len(self.sourcelines):
            l = 1
        while not self.sourcelines[l - 1]:
            l += 1
            if l > len(self.sourcelines):
                break
        return l

    @property
    def current_line(self):
        try:
            return self.sourcelines[self.lineno - 1]
        except IndexError:
            return u''

    @cachedprop
    def render_locals(self):
        rv = dict()
        for k, v in iteritems(self.locals):
            try:
                rv[k] = str(v)
            except:
                rv[k] = '<unavailable>'
        return rv


def make_traceback(exc_info, template_ref=None):
    if template_ref:
        if isinstance(exc_info[1], SyntaxError):
            if exc_info[1].filename == '<string>':
                exc_info = translate_syntax_error(exc_info[1], template_ref)
    if isinstance(exc_info[1], TemplateError):
        exc_info = translate_template_error(exc_info[1])
    tb = translate_exception(exc_info, 1)
    exc_type, exc_value, tb = tb.standard_exc_info
    reraise(exc_type, exc_value, tb)


def translate_syntax_error(error, reference):
    """Rewrites a syntax error to please traceback systems."""
    #error.translated = True
    #exc_info = (error.__class__, error, None)
    exc_info = (error.__class__, 'invalid syntax', None)
    filename = reference.file_path
    if filename is None:
        filename = '<unknown>'
    return fake_exc_info(exc_info, filename, reference.lineno)


def translate_template_error(error):
    exc_info = (error.__class__, error, None)
    filename = error.file_path
    if filename is None:
        filename = '<unknown>'
    return fake_exc_info(exc_info, filename, error.lineno)


def smart_traceback(app):
    exc_type, exc_value, tb = sys.exc_info()
    return Traceback(app, exc_type, exc_value, tb)


def debug_handler(tb):
    from os.path import join, dirname, basename
    view = join(dirname(__file__), 'assets', 'debug', basename('view.html'))
    from .templating.core import DummyResponse
    from .templating.parser import TemplateParser
    view_file = open(view, 'rb')
    view_source = view_file.read().decode('utf8')
    view_file.close()
    context = {'_DummyResponse_': DummyResponse(), 'tb': tb}
    from .datastructures import sdict
    t_dict = sdict(lexers={})
    code = str(TemplateParser(t_dict, view_source,
               context=context, path=''))
    exec(code, context)
    return context['_DummyResponse_'].body.getvalue()


def make_frame_proxy(frame):
    proxy = TracebackFrameProxy(frame)
    if tproxy is None:
        return proxy

    def operation_handler(operation, *args, **kwargs):
        if operation in ('__getattribute__', '__getattr__'):
            return getattr(proxy, args[0])
        elif operation == '__setattr__':
            proxy.__setattr__(*args, **kwargs)
        else:
            return getattr(proxy, operation)(*args, **kwargs)
    return tproxy(TracebackType, operation_handler)


def translate_exception(exc_info, initial_skip=0):
    """If passed an exc_info it will automatically rewrite the exceptions
    all the way down to the correct line numbers and frames.
    """
    tb = exc_info[2]
    frames = []

    # skip some internal frames if wanted
    for x in range(initial_skip):
        if tb is not None:
            tb = tb.tb_next

    while tb is not None:
        # save a reference to the next frame if we override the current
        # one with a faked one.
        next = tb.tb_next

        # fake template exceptions
        template = tb.tb_frame.f_globals.get('__weppy_template__')
        if template is not None:
            lineno = template.lineno
            tb = fake_exc_info(exc_info[:2] + (tb,), template.file_path,
                               lineno)[2]

        frames.append(make_frame_proxy(tb))
        tb = next

    # if we don't have any exceptions in the frames left, we have to
    # reraise it unchanged.
    # XXX: can we backup here?  when could this happen?
    if not frames:
        reraise(exc_info[0], exc_info[1], exc_info[2])

    return ProcessedTraceback(exc_info[0], exc_info[1], frames)


def fake_exc_info(exc_info, filename, lineno):
    """Helper for `translate_exception`."""
    exc_type, exc_value, tb = exc_info

    # figure the real context out
    if tb is not None:
        real_locals = tb.tb_frame.f_locals.copy()
        ctx = real_locals.get('context')
        if ctx:
            locals = ctx
        else:
            locals = {}
        #for name, value in iteritems(real_locals):
        #    if name.startswith('l_') and value is not missing:
        #        locals[name[2:]] = value

        # if there is a local called __weppy_exception__, we get
        # rid of it to not break the debug functionality.
        locals.pop('__weppy_exception__', None)
    else:
        locals = {}

    # assamble fake globals we need
    globals = {
        '__name__': filename,
        '__file__': filename,
        '__weppy_exception__': exc_info[:2],

        # we don't want to keep the reference to the template around
        # to not cause circular dependencies, but we mark it as weppy
        # frame for the ProcessedTraceback
        '__weppy_template__': None
    }

    # and fake the exception
    code = compile('\n' * (lineno - 1) + raise_helper, filename, 'exec')

    # if it's possible, change the name of the code.  This won't work
    # on some python environments such as google appengine
    try:
        code = CodeType(0, code.co_nlocals, code.co_stacksize,
                        code.co_flags, code.co_code, code.co_consts,
                        code.co_names, code.co_varnames, filename,
                        'template', code.co_firstlineno,
                        code.co_lnotab, (), ())
    except:
        pass

    # execute the code and catch the new traceback
    try:
        exec(code, globals, locals)
    except:
        exc_info = sys.exc_info()
        new_tb = exc_info[2].tb_next

    # return without this frame
    return exc_info[:2] + (new_tb,)


def _init_ugly_crap():
    """This function implements a few ugly things so that we can patch the
    traceback objects.  The function returned allows resetting `tb_next` on
    any python traceback object.  Do not attempt to use this on non cpython
    interpreters
    """
    import ctypes

    # figure out side of _Py_ssize_t
    if hasattr(ctypes.pythonapi, 'Py_InitModule4_64'):
        _Py_ssize_t = ctypes.c_int64
    else:
        _Py_ssize_t = ctypes.c_int

    # regular python
    class _PyObject(ctypes.Structure):
        pass
    _PyObject._fields_ = [
        ('ob_refcnt', _Py_ssize_t),
        ('ob_type', ctypes.POINTER(_PyObject))
    ]

    # python with trace
    if hasattr(sys, 'getobjects'):
        class _PyObject(ctypes.Structure):
            pass
        _PyObject._fields_ = [
            ('_ob_next', ctypes.POINTER(_PyObject)),
            ('_ob_prev', ctypes.POINTER(_PyObject)),
            ('ob_refcnt', _Py_ssize_t),
            ('ob_type', ctypes.POINTER(_PyObject))
        ]

    class _Traceback(_PyObject):
        pass
    _Traceback._fields_ = [
        ('tb_next', ctypes.POINTER(_Traceback)),
        ('tb_frame', ctypes.POINTER(_PyObject)),
        ('tb_lasti', ctypes.c_int),
        ('tb_lineno', ctypes.c_int)
    ]

    def tb_set_next(tb, next):
        """Set the tb_next attribute of a traceback object."""
        if not (isinstance(tb, TracebackType) and
                (next is None or isinstance(next, TracebackType))):
            raise TypeError('tb_set_next arguments must be traceback objects')
        obj = _Traceback.from_address(id(tb))
        if tb.tb_next is not None:
            old = _Traceback.from_address(id(tb.tb_next))
            old.ob_refcnt -= 1
        if next is None:
            obj.tb_next = ctypes.POINTER(_Traceback)()
        else:
            next = _Traceback.from_address(id(next))
            next.ob_refcnt += 1
            obj.tb_next = ctypes.pointer(next)

    return tb_set_next


# try to get a tb_set_next implementation if we don't have transparent
# proxies.
tb_set_next = None
if tproxy is None:
    try:
        tb_set_next = _init_ugly_crap()
    except:
        pass
    del _init_ugly_crap



# -*- coding: utf-8 -*-
"""
    weppy.language.translator
    -------------------------

    The translator main logic.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's translation system (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import os

from .._compat import PY2, implements_bool, implements_to_string, iteritems, \
    iterkeys, to_unicode
from ..tags import asis, htmlescape
from ..utils import cachedprop
from .helpers import regex_backslash, regex_plural, regex_plural_dict, \
    regex_plural_tuple, regex_language, DEFAULT_NPLURALS, \
    DEFAULT_GET_PLURAL_ID, DEFAULT_CONSTRUCT_PLURAL_FORM, \
    read_possible_languages, read_dict, write_dict, read_plural_dict, \
    write_plural_dict, ttab_in, ttab_out, upper_fun, title_fun, cap_fun
from .cache import get_from_cache

if PY2:
    NUMBERS = (int, long, float)
else:
    NUMBERS = (int, float)


#: The single 'translator string element', is created when user calls
#  T('string'), and will be translated when loaded in templates or converted to
#  a string (via str())
@implements_bool
@implements_to_string
class TElement(object):
    m = s = T = language = None
    M = is_copy = False

    def __init__(self, T, message, symbols={}, M=False, language=None):
        if isinstance(message, TElement):
            self.m = message.m
            self.s = message.s
            self.T = message.T
            self.M = message.M
            self.language = message.language
            self.is_copy = True
        else:
            self.m = message
            self.s = symbols
            self.T = T
            self.M = M
            self.language = language
            self.is_copy = False

    def __repr__(self):
        return "<lazyT %s>" % repr(self.m)

    def __str__(self):
        lang = self.language
        if lang is None:
            #: use language provided by http_accept_language or
            #  url (if forced by application), fallback on default language
            from ..globals import current
            lang = current._language or self.T.current_languages[0]
        # return str(self.T.apply_filter(lang, self.m, self.s) if self.M else
        #            self.T.translate(lang, self.m, self.s))
        return self.T.translate(lang, self.m, self.s)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def __add__(self, other):
        return '%s%s' % (self, other)

    def __radd__(self, other):
        return '%s%s' % (other, self)

    def __mul__(self, other):
        return str(self) * other

    def __cmp__(self, other):
        return cmp(str(self), str(other))

    def __hash__(self):
        return hash(str(self))

    def __getattr__(self, name):
        return getattr(str(self), name)

    def __getitem__(self, i):
        return str(self)[i]

    def __getslice__(self, i, j):
        return str(self)[i:j]

    def __iter__(self):
        for c in str(self):
            yield c

    def __len__(self):
        return len(str(self))

    def __bool__(self):
        return len(self.m) > 0

    # def to_html(self):
    #     return str(self) if self.M else escape(str(self))

    def encode(self, *a, **b):
        return str(self).encode(*a, **b)

    def decode(self, *a, **b):
        return str(self).decode(*a, **b)

    def read(self):
        return str(self)

    def __mod__(self, symbols):
        if self.is_copy:
            return TElement(self)
        return TElement(self.T, self.m, symbols, self.M)


#: A language for the translator.
class TLanguage(object):
    def __init__(self, translator, l_info, filename=None, writable=True):
        self.translator = translator
        self.filename = filename
        self.writable = writable
        self.cached = translator.cached
        self._init_plurals(l_info)

    def _init_plurals(self, lang_info):
        if lang_info:
            (pname, pmtime, self.plural_language, self.nplurals,
             self.get_plural_id, self.construct_plural_form) = lang_info[3:]
            self.pmtime = pmtime
            if pname:
                pname = os.path.join(self.translator.langpath, pname)
            self.plural_file = pname
        else:
            self.plural_language = 'default'
            self.nplurals = DEFAULT_NPLURALS
            self.get_plural_id = DEFAULT_GET_PLURAL_ID
            self.construct_plural_form = DEFAULT_CONSTRUCT_PLURAL_FORM
            self.pmtime = 0
            self.plural_file = None

    def _read_t(self):
        if self.filename is None:
            return {}
        return read_dict(self.filename)

    @cachedprop
    def _t(self):
        return self._read_t()

    @property
    def t(self):
        if self.cached:
            return self._t
        return self._read_t()

    def set(self, key, message):
        # [WARN]: if it's not cached, is safe?
        self.t[key] = message
        if self.writable and self.filename:
            write_dict(self.filename, self.t)

    def get(self, key):
        return self.t.get(key)

    def get_default(self, key, message):
        if self.translator.default_translator is self:
            return message
        return self.translator.default_translator.t.get(key, message)

    def _read_pl(self):
        if self.plural_file is None or self.pmtime == 0:
            return {}
        return read_plural_dict(self.plural_file)

    @cachedprop
    def _pl(self):
        return self._read_pl()

    @property
    def pl(self):
        if self.cached:
            return self._pl
        return self._read_pl()

    def plural(self, word, n):
        """ get plural form of word for number *n*
        args:
            word (str): word in singular
            n (numeric): number plural form created for

        returns:
            (str): word in appropriate singular/plural form
        """
        if int(n) == 1:
            return word
        elif word:
            id = self.get_plural_id(abs(int(n)))
            # id = 0 singular form
            # id = 1 first plural form
            # id = 2 second plural form
            # etc.
            if id != 0:
                forms = self.pl.get(word, [])
                if len(forms) >= id:
                    # have this plural form:
                    return forms[id - 1]
                else:
                    # guessing this plural form
                    forms += [''] * (self.nplurals - len(forms) - 1)
                    form = self.construct_plural_form(word, id)
                    forms[id - 1] = form
                    self.pl[word] = forms
                    if self.writable and self.plural_file:
                        write_plural_dict(self.plural_file,
                                          self.pl)
                    return form
        return word

    def get_t(self, message, prefix=u''):
        """
        use ## to add a comment into a translation string
        the comment can be useful do discriminate different possible
        translations for the same string (for example different locations)

        T(' hello world ') -> ' hello world '
        T(' hello world ## token') -> ' hello world '
        T('hello ## world## token') -> 'hello ## world'

        the ## notation is ignored in multiline strings and strings that
        start with ##. this is to allow markmin syntax to be translated
        """
        message = to_unicode(message)
        key = prefix + message
        mt = self.get(key)
        if mt is not None:
            return mt
        # we did not find a translation
        if message.find('##') > 0 and '\n' not in message:
            # remove comments
            message = message.rsplit('##', 1)[0]
        # guess translation same as original
        #self.t[key] = mt = self.default_t.get(key, message)
        mt = self.get_default(key, message)
        self.set(key, mt)
        return regex_backslash.sub(
            lambda m: m.group(1).translate(ttab_in), mt)

    def translate(self, message):
        if self.cached:
            return get_from_cache(
                self.filename, to_unicode(message),
                lambda: to_unicode(self.get_t(message)))
        return to_unicode(self.get_t(message))


#: The main translator object, responsible of creating elements and loading
#  application languages
class Translator(object):
    def __init__(self, app):
        self.app = app
        self.langpath = os.path.join(app.root_path, "languages")
        #self.is_writable = not is_gae
        self.is_writable = app.language_write
        def_langs = app.language_default if \
            isinstance(app.language_default, (tuple, list)) \
            else [app.language_default]
        self.TLanguages = {}
        self.set_current_languages(def_langs)
        #self.filter = markmin
        #self.ftag = 'markmin'

    @property
    def cached(self):
        return not self.app.debug

    @property
    def possible_languages(self):
        """
        return info for selected language or dictionary with all
            possible languages info from languages/*.py
        """
        #: if translator is cached, it doesn't check for filesystem updates
        if not self.cached:
            return read_possible_languages(self.langpath)
        if not hasattr(self, '_pl_info'):
            self._pl_info = read_possible_languages(self.langpath)
        return self._pl_info

    def get_language_info(self, lang):
        return self.possible_languages.get(lang)

    def set_current_languages(self, languages):
        """
        set current AKA "default" languages
        setting one of this languages makes force() function
        turn translation off to use default language
        """
        #: get default language info from default.py/DEFAULT_LANGUAGE
        pl_info = self.get_language_info('default')
        #: pl_info[2] is the langfile_mtime, if it equals 0 it means
        #  app doesn't have a languages/default.py
        if pl_info[2] == 0:
            self.default_language_file = self.langpath
        else:
            self.default_language_file = os.path.join(self.langpath,
                                                      'default.py')
        #: set default languages
        self.current_languages = [pl_info[0]]
        self.add_translator(pl_info[0], self.default_language_file, True)
        self.default_translator = self.TLanguages[pl_info[0]]
        for lang in languages:
            if lang not in self.current_languages:
                self.TLanguages[lang] = self.default_translator
                self.current_languages.append(lang)

    def add_translator(self, language, filename, default=False):
        writable = not default and self.is_writable
        pl_info = self.get_language_info(language)
        if filename == self.langpath:
            self.TLanguages[language] = TLanguage(self, pl_info,
                                                  writable=writable)
        else:
            self.TLanguages[language] = TLanguage(self, pl_info,
                                                  filename=filename,
                                                  writable=writable)

    def build_translator(self, language):
        all_languages = set(lang for lang in iterkeys(self.possible_languages)
                            if lang != 'default')
        # compare "aa-bb" | "aa" from *language* parameter
        # with strings from langlist using such alghorythm:
        # xx-yy.py -> xx.py -> xx*.py
        found = None
        lang5 = language[:5]
        if lang5 in all_languages:
            found = lang5
        else:
            lang2 = language[:2]
            if len(lang5) > 2 and lang2 in all_languages:
                found = lang2
            else:
                for l in all_languages:
                    if l[:2] == lang2:
                        found = l
        if found and found not in self.current_languages:
            #: if we can't find a translator instance for the found lang,
            #  we build it
            if self.TLanguages.get(found) is None:
                filename = os.path.join(self.langpath, found + '.py')
                self.add_translator(found, filename)
            #: we map language with found (example: `en-us` and `en` with only
            #  en.py file)
            if language != found:
                self.TLanguages[language] = self.TLanguages[found]
        else:
            self.TLanguages[language] = self.default_translator

    def __call__(self, message, symbols={}, language=None):
        return TElement(self, message, symbols, language=language)

    def apply_filter(self, message, symbols={}, filter=None, ftag=None):
        def get_tr(message, prefix, filter):
            s = self.get_t(message, prefix)
            return filter(s) if filter else self.filter(s)
        if filter:
            prefix = '@' + (ftag or 'userdef') + '\x01'
        else:
            prefix = '@' + self.ftag + '\x01'
        message = get_from_cache(
            self.cache, prefix + message,
            lambda: get_tr(message, prefix, filter))
        if symbols or symbols == 0 or symbols == "":
            if isinstance(symbols, dict):
                symbols.update(
                    (key, htmlescape(value).translate(ttab_in))
                    for key, value in iteritems(symbols)
                    if not isinstance(value, NUMBERS))
            else:
                if not isinstance(symbols, tuple):
                    symbols = (symbols,)
                symbols = tuple(
                    value if isinstance(value, NUMBERS)
                    else htmlescape(value).translate(ttab_in)
                    for value in symbols)
            message = self.params_substitution(message, symbols)
        return asis(message.translate(ttab_out))

    #def M(self, message, symbols={}, language=None,
    #      lazy=None, filter=None, ftag=None):
    #    """
    #    get cached translated markmin-message with inserted parametes
    #    if lazy==True lazyT object is returned
    #    """
    #    if lazy is None:
    #        lazy = self.lazy
    #    if not language:
    #        if lazy:
    #            return lazyT(message, symbols, self, filter, ftag, True)
    #        else:
    #            return self.apply_filter(message, symbols, filter, ftag)
    #    else:
    #        try:
    #            otherT = self.otherTs[language]
    #        except KeyError:
    #            otherT = self.otherTs[language] = translator(self.request)
    #            otherT.force(language)
    #        return otherT.M(message, symbols, lazy=lazy)

    def params_substitution(self, lang, message, symbols):
        """
        substitute parameters from symbols into message using %.
        also parse %%{} placeholders for plural-forms processing.
        returns: string with parameters
        NOTE: *symbols* MUST BE OR tuple OR dict of parameters!
        """
        def sub_plural(m):
            """string in %{} is transformed by this rules:
               If string starts with  \\, ! or ? such transformations
               take place:

               "!string of words" -> "String of word" (Capitalize)
               "!!string of words" -> "String Of Word" (Title)
               "!!!string of words" -> "STRING OF WORD" (Upper)
               "\\!string of words" -> "!string of word"
                             (remove \\ and disable transformations)
               "?word?number" -> "word" (return word, if number == 1)
               "?number" or "??number" -> "" (remove number,
                                              if number == 1)
               "?word?number" -> "number" (if number != 1)
            """
            def sub_tuple(m):
                w, i = m.group('w', 'i')
                c = w[0]
                if c not in '!?':
                    return self.TLanguages[lang].plural(w,
                                                        symbols[int(i or 0)])
                elif c == '?':
                    (p1, sep, p2) = w[1:].partition("?")
                    part1 = p1 if sep else ""
                    (part2, sep, part3) = (p2 if sep else p1).partition("?")
                    if not sep:
                        part3 = part2
                    if i is None:
                        # ?[word]?number[?number] or ?number
                        if not part2:
                            return m.group(0)
                        num = int(part2)
                    else:
                        # ?[word]?word2[?word3][number]
                        num = int(symbols[int(i or 0)])
                    return part1 if num == 1 else part3 if num == 0 else part2
                elif w.startswith('!!!'):
                    word = w[3:]
                    fun = upper_fun
                elif w.startswith('!!'):
                    word = w[2:]
                    fun = title_fun
                else:
                    word = w[1:]
                    fun = cap_fun
                if i is not None:
                    return fun(self.TLanguages[lang].plural(word,
                               symbols[int(i)]))
                return fun(word)

            def sub_dict(m):
                w, n = m.group('w', 'n')
                c = w[0]
                n = int(n) if n.isdigit() else symbols[n]
                if c not in '!?':
                    return self.TLanguages[lang].plural(w, n)
                elif c == '?':
                    (p1, sep, p2) = w[1:].partition("?")
                    part1 = p1 if sep else ""
                    (part2, sep, part3) = (p2 if sep else p1).partition("?")
                    if not sep:
                        part3 = part2
                    num = int(n)
                    return part1 if num == 1 else part3 if num == 0 else part2
                elif w.startswith('!!!'):
                    word = w[3:]
                    fun = upper_fun
                elif w.startswith('!!'):
                    word = w[2:]
                    fun = title_fun
                else:
                    word = w[1:]
                    fun = cap_fun
                return fun(self.TLanguages[lang].plural(word, n))

            s = m.group(1)
            part = regex_plural_tuple.sub(sub_tuple, s)
            if part == s:
                part = regex_plural_dict.sub(sub_dict, s)
                if part == s:
                    return m.group(0)
            return part
        message = message % symbols
        message = regex_plural.sub(sub_plural, message)
        return message

    def get_best_language(self, langstr):
        langs = regex_language.findall(langstr)
        for lang in langs:
            if lang not in self.current_languages and \
                    lang[:2] not in self.current_languages:
                return lang
        return langs[0] if langs else self.current_languages[0]

    def translate(self, lang, message, symbols):
        lang = self.get_best_language(lang.lower())
        if self.TLanguages.get(lang) is None:
            self.build_translator(lang)
        message = self.TLanguages[lang].translate(message)
        if symbols or symbols == 0 or symbols == "":
            if isinstance(symbols, dict):
                symbols.update(
                    (
                        key, to_unicode(value).translate(ttab_in)
                        if value else u'')
                    for key, value in iteritems(symbols)
                    if not isinstance(value, NUMBERS))
            else:
                if not isinstance(symbols, tuple):
                    symbols = (symbols,)
                symbols = tuple(
                    value if isinstance(value, NUMBERS)
                    else (
                        to_unicode(value).translate(ttab_in) if value else u'')
                    for value in symbols)
            message = self.params_substitution(lang, message, symbols)
        return message.translate(ttab_out)



from ...datastructures import sdict
from ...http import redirect
from .models import AuthUser, AuthGroup, AuthMembership, AuthPermission, \
    AuthEvent


DEFAULT = lambda: None

default_settings = sdict(
    password_min_length=6,
    cas_maps=None,
    registration_requires_verification=False,
    registration_requires_approval=False,
    login_after_registration=False,
    login_after_password_change=True,
    create_user_groups=None,
    everybody_group_id=None,
    #login_captcha=None,
    #register_captcha=None,
    #retrieve_username_captcha=None,
    #retrieve_password_captcha=None,
    #captcha=None,
    expiration=3600,  # one hour
    long_expiration=3600 * 30 * 24,  # one month
    remember_me_form=True,
    allow_basic_login=False,
    on_failed_authentication=lambda x: redirect(x),
    logging_enabled=True,
    allow_delete_accounts=False,
    #password_field='password',
    #table_user_name='auth_user',
    #table_group_name='auth_group',
    #table_membership_name='auth_membership',
    #table_permission_name='auth_permission',
    #table_event_name='auth_event',
    #table_cas_name='auth_cas',
    models=sdict(
        user=AuthUser,
        group=AuthGroup,
        membership=AuthMembership,
        permission=AuthPermission,
        event=AuthEvent
    ),
    use_username=False,
    login_userfield='email',
    logout_onlogout=None,
    register_fields=None,
    profile_fields=None,
    email_case_sensitive=True,
    username_case_sensitive=True,
    update_fields=['email'],
    ondelete="CASCADE",
    extra_fields={},
    actions_disabled=[],
    login_onvalidation=[],
    login_onaccept=[],
    login_onfail=[],
    register_onvalidation=[],
    register_onaccept=[],
    verify_email_onaccept=[],
    profile_onvalidation=[],
    profile_onaccept=[],
    change_password_onvalidation=[],
    change_password_onaccept=[],
    retrieve_password_onvalidation=[],
    reset_password_onvalidation=[],
    reset_password_onaccept=[],
    messages=sdict()
)

default_messages = sdict(
    login_button='Login',
    register_button='Register',
    password_reset_button='Request reset password',
    password_change_button='Change password',
    profile_save_button='Apply changes',
    submit_button='Submit',
    verify_password='Verify Password',
    delete_label='Check to delete',
    function_disabled='Function disabled',
    access_denied='Insufficient privileges',
    registration_verifying='Registration needs verification',
    registration_pending='Registration is pending approval',
    email_taken='This email already has an account',
    invalid_username='Invalid username',
    username_taken='Username already taken',
    login_disabled='Login disabled by administrator',
    logged_in='Logged in',
    email_sent='Email sent',
    unable_to_send_email='Unable to send email',
    email_verified='Email verified',
    logged_out='Logged out',
    registration_successful='Registration successful',
    invalid_email='Invalid email',
    unable_send_email='Unable to send email',
    invalid_login='Invalid login',
    invalid_user='Invalid user',
    invalid_password='Invalid password',
    is_empty="Cannot be empty",
    mismatched_password="Password fields don't match",
    verify_email='Welcome %(username)s! Click on the link %(link)s to verify your email',
    verify_email_subject='Email verification',
    username_sent='Your username was emailed to you',
    new_password_sent='A new password was emailed to you',
    password_changed='Password changed',
    retrieve_username='Your username is: %(username)s',
    retrieve_username_subject='Username retrieve',
    retrieve_password='Your password is: %(password)s',
    retrieve_password_subject='Password retrieve',
    reset_password='Click on the link %(link)s to reset your password',
    reset_password_subject='Password reset',
    invalid_reset_password='Invalid reset password',
    profile_updated='Profile updated',
    new_password='New password',
    old_password='Old password',
    group_description='Group uniquely assigned to user %(id)s',
    register_log='User %(id)s Registered',
    login_log='User %(id)s Logged-in',
    login_failed_log=None,
    logout_log='User %(id)s Logged-out',
    profile_log='User %(id)s Profile updated',
    verify_email_log='User %(id)s Verification email sent',
    retrieve_username_log='User %(id)s Username retrieved',
    retrieve_password_log='User %(id)s Password retrieved',
    reset_password_log='User %(id)s Password reset',
    change_password_log='User %(id)s Password changed',
    add_group_log='Group %(group_id)s created',
    del_group_log='Group %(group_id)s deleted',
    add_membership_log=None,
    del_membership_log=None,
    has_membership_log=None,
    add_permission_log=None,
    del_permission_log=None,
    has_permission_log=None,
    impersonate_log='User %(id)s is impersonating %(other_id)s',
    label_first_name='First name',
    label_last_name='Last name',
    label_username='Username',
    label_email='E-mail',
    label_password='Password',
    label_registration_key='Registration key',
    label_reset_password_key='Reset Password key',
    label_registration_id='Registration identifier',
    label_role='Role',
    label_description='Description',
    label_user_id='User ID',
    label_group_id='Group ID',
    label_name='Name',
    label_table_name='Object or table name',
    label_record_id='Record ID',
    label_time_stamp='Timestamp',
    label_client_ip='Client IP',
    label_origin='Origin',
    label_remember_me="Remember me (for 30 days)",
    verify_password_comment='please input your password again',
)



# -*- coding: utf-8 -*-
"""
    weppy.helpers
    -------------

    Provides helping methods for applications.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""
from ._compat import string_types


def abort(code, body=''):
    from .http import HTTP
    from .globals import current
    current.response.status = code
    raise HTTP(code, body)


def stream_file(path):
    import os
    from .globals import request, response
    from .expose import Expose
    from .stream import streamer
    fullfilename = os.path.join(Expose.application.root_path, path)
    raise streamer(request.environ, fullfilename, headers=response.headers)


def stream_dbfile(db, name):
    import re
    items = re.compile('(?P<table>.*?)\.(?P<field>.*?)\..*').match(name)
    if not items:
        abort(404)
    (t, f) = (items.group('table'), items.group('field'))
    try:
        field = db[t][f]
    except AttributeError:
        abort(404)
    from pydal.exceptions import NotAuthorizedException, NotFoundException
    try:
        (filename, fullfilename) = field.retrieve(name, nameonly=True)
    except NotAuthorizedException:
        abort(403)
    except NotFoundException:
        abort(404)
    except IOError:
        abort(404)
    from .globals import request, response
    if isinstance(fullfilename, string_types):
        #: handle file uploads
        from .stream import streamer
        raise streamer(request.environ, fullfilename, headers=response.headers)
    else:
        #: handle blob fields
        from .libs.contenttype import contenttype
        from .http import HTTP
        response.headers['Content-Type'] = contenttype(filename)
        if 'wsgi.file_wrapper' in request.environ:
            data = request.environ['wsgi.file_wrapper'](fullfilename, 10**5)
        else:
            data = iter(lambda: fullfilename.read(10**5), '')
        raise HTTP(200, data, response.headers)


def flash(message, category='message'):
    #: Flashes a message to the next request.
    from .globals import session
    if session._flashes is None:
        session._flashes = []
    session._flashes.append((category, message))


def get_flashed_messages(with_categories=False, category_filter=[]):
    #: Pulls flashed messages from the session and returns them.
    #  By default just the messages are returned, but when `with_categories`
    #  is set to `True`, the return value will be a list of tuples in the
    #  form `(category, message)` instead.
    from .globals import session
    try:
        flashes = list(session._flashes or [])
        if category_filter:
            flashes = list(filter(lambda f: f[0] in category_filter, flashes))
        for el in flashes:
            session._flashes.remove(el)
        if not with_categories:
            return [x[1] for x in flashes]
    except:
        flashes = []
    return flashes


def load_component(url, target=None, content='loading...'):
    from .tags import tag
    attr = {}
    if target:
        attr['_id'] = target
    attr['_data-wpp_remote'] = url
    return tag.div(content, **attr)



# -*- coding: utf-8 -*-
"""
    weppy.logger
    ------------

    Provides logging utitilites for weppy applications.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the code of Flask (http://flask.pocoo.org)
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import
import os
from threading import Lock
from logging import getLogger, StreamHandler, Formatter, getLoggerClass, DEBUG
from logging.handlers import RotatingFileHandler
from .datastructures import sdict

_logger_lock = Lock()

_levels = {
    'debug': 10,
    'info': 20,
    'warning': 30,
    'error': 40,
    'critical': 50
}

_def_log_config = sdict(
    production=sdict(
        max_size=5 * 1024 * 1024,
        file_no=4,
        level='warning',
        format='[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        on_app_debug=False))

_debug_log_format = (
    '> %(levelname)s in %(module)s [%(pathname)s:%(lineno)d]:\n' +
    '%(message)s'
)


def create_logger(app):
    Logger = getLoggerClass()

    class DebugLogger(Logger):
        def getEffectiveLevel(x):
            if x.level == 0 and app.debug:
                return DEBUG
            return Logger.getEffectiveLevel(x)

    class DebugHandler(StreamHandler):
        def emit(x, record):
            StreamHandler.emit(x, record) if app.debug else None

    class DebugRFHandler(RotatingFileHandler):
        def emit(x, record):
            RotatingFileHandler.emit(x, record) if app.debug else None

    class ProdRFHandler(RotatingFileHandler):
        def emit(x, record):
            RotatingFileHandler.emit(x, record) if not app.debug else None

    # init the console debug handler
    debug_handler = DebugHandler()
    debug_handler.setLevel(DEBUG)
    debug_handler.setFormatter(Formatter(_debug_log_format))
    logger = getLogger(app.logger_name)
    # just in case that was not a new logger, get rid of all the handlers
    # already attached to it.
    del logger.handlers[:]
    logger.__class__ = DebugLogger
    logger.addHandler(debug_handler)
    # load application logging config
    app_logs = app.config.logging
    if not app_logs:
        app_logs = _def_log_config
    for lname, lconf in app_logs.items():
        lfile = os.path.join(app.root_path, 'logs', lname + '.log')
        max_size = lconf.max_size or _def_log_config.production.max_size
        file_no = lconf.file_no or _def_log_config.production.file_no
        level = _levels.get(lconf.level or 'warning', _levels.get('warning'))
        lformat = lconf.format or _def_log_config.production.format
        on_app_debug = lconf.on_app_debug
        if on_app_debug:
            handler = DebugRFHandler(lfile, maxBytes=max_size,
                                     backupCount=file_no)
        else:
            handler = ProdRFHandler(lfile, maxBytes=max_size,
                                    backupCount=file_no)
        handler.setLevel(level)
        handler.setFormatter(Formatter(lformat))
        logger.addHandler(handler)
    return logger



# -*- coding: utf-8 -*-
"""
    weppy._reloader
    ---------------

    Provides auto-reloading utilities.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Adapted from werkzeug code (http://werkzeug.pocoo.org)
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

from __future__ import print_function
import os
import sys
import time
import subprocess
import threading
from itertools import chain

from ._compat import PY2, iteritems


def _iter_module_files(reloader_paths=None):
    """This iterates over all relevant Python files.  It goes through all
    loaded files from modules, all files in folders of already loaded modules
    as well as all files reachable through a package.
    """
    found = set()
    entered = set()

    def _verify_file(filename):
        if not filename:
            return
        filename = os.path.abspath(filename)
        old = None
        while not os.path.isfile(filename):
            old = filename
            filename = os.path.dirname(filename)
            if filename == old:
                break
        else:
            if filename[-4:] in ('.pyc', '.pyo'):
                filename = filename[:-1]
            if filename not in found:
                found.add(filename)
                return filename

    def _recursive_walk(path_entry):
        if path_entry in entered:
            return
        entered.add(path_entry)
        try:
            for filename in os.listdir(path_entry):
                path = os.path.join(path_entry, filename)
                if os.path.isdir(path):
                    for filename in _recursive_walk(path):
                        yield filename
                else:
                    if not filename.endswith(('.py', '.pyc', '.pyo')):
                        continue
                    filename = _verify_file(path)
                    if filename:
                        yield filename
        except OSError:
            pass

    if not reloader_paths:
        # The list call is necessary on Python 3 in case the module
        # dictionary modifies during iteration.
        for path_entry in list(sys.path):
            for filename in _recursive_walk(os.path.abspath(path_entry)):
                yield filename

        for module in list(sys.modules.values()):
            if module is None:
                continue
            filename = _verify_file(getattr(module, '__file__', None))
            if filename:
                yield filename
                for filename in _recursive_walk(os.path.dirname(filename)):
                    yield filename
            for package_path in getattr(module, '__path__', ()):
                for filename in _recursive_walk(os.path.abspath(package_path)):
                    yield filename
    else:
        for path_entry in reloader_paths:
            for filename in _recursive_walk(os.path.abspath(path_entry)):
                yield filename


def _find_common_roots(paths):
    """Out of some paths it finds the common roots that need monitoring."""
    paths = [x.split(os.path.sep) for x in paths]
    root = {}
    for chunks in sorted(paths, key=len, reverse=True):
        node = root
        for chunk in chunks:
            node = node.setdefault(chunk, {})
        node.clear()

    rv = set()

    def _walk(node, path):
        for prefix, child in iteritems(node):
            _walk(child, path + (prefix,))
        if not node:
            rv.add('/'.join(path))
    _walk(root, ())
    return rv


class ReloaderLoop(object):
    name = None

    # monkeypatched by testsuite. wrapping with `staticmethod` is required in
    # case time.sleep has been replaced by a non-c function (e.g. by
    # `eventlet.monkey_patch`) before we get here
    _sleep = staticmethod(time.sleep)

    def __init__(self, reloader_paths=None, extra_files=None, interval=1):
        self.reloader_paths = reloader_paths
        self.extra_files = set(os.path.abspath(x)
                               for x in extra_files or ())
        self.interval = interval

    def run(self):
        pass

    def restart_with_reloader(self):
        """Spawn a new Python interpreter with the same arguments as this one,
        but running the reloader thread.
        """
        while 1:
            #self.log('> Restarting (%s mode)' % self.name)
            print('> Restarting (%s mode)' % self.name)
            args = [sys.executable] + sys.argv
            new_environ = os.environ.copy()
            new_environ['WEPPY_RUN_MAIN'] = 'true'

            # a weird bug on windows. sometimes unicode strings end up in the
            # environment and subprocess.call does not like this, encode them
            # to latin1 and continue.
            if os.name == 'nt' and PY2:
                for key, value in iteritems(new_environ):
                    if isinstance(value, unicode):
                        new_environ[key] = value.encode('iso-8859-1')

            exit_code = subprocess.call(args, env=new_environ)
            if exit_code != 3:
                return exit_code

    def trigger_reload(self, filename):
        filename = os.path.abspath(filename)
        #self.log('> Detected change in %r, reloading' % filename)
        print('> Detected change in %r, reloading' % filename)
        sys.exit(3)


class StatReloaderLoop(ReloaderLoop):
    name = 'stat'

    def run(self):
        mtimes = {}
        while 1:
            for filename in chain(_iter_module_files(self.reloader_paths),
                                  self.extra_files):
                try:
                    mtime = os.stat(filename).st_mtime
                except OSError:
                    continue

                old_time = mtimes.get(filename)
                if old_time is None:
                    mtimes[filename] = mtime
                    continue
                elif mtime > old_time:
                    self.trigger_reload(filename)
            self._sleep(self.interval)


reloader_loops = {
    'stat': StatReloaderLoop,
}

reloader_loops['auto'] = reloader_loops['stat']


def run_with_reloader(app, host, port, extra_files=None, interval=1,
                      reloader_type='auto', reloader_paths=None):
    """Run the given function in an independent python interpreter."""
    import signal
    reloader = reloader_loops[reloader_type](reloader_paths, extra_files,
                                             interval)
    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
    try:
        if os.environ.get('WEPPY_RUN_MAIN') == 'true':
            t = threading.Thread(target=app._run, args=(host, port))
            t.setDaemon(True)
            t.start()
            reloader.run()
        else:
            sys.exit(reloader.restart_with_reloader())
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
"""
    weppy.testing.client
    --------------------

    Provides base classes for testing suite.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Several parts of this code comes from Werkzeug.
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

import copy
from itertools import chain
from .._compat import reraise, text_type, to_native
from .._internal import ClosingIterator
from ..utils import cachedprop
from .env import EnvironBuilder
from .helpers import TestCookieJar, Headers
from .urls import get_host, url_parse, url_unparse


class ClientContext(object):
    def __init__(self):
        from ..globals import current, Request
        self.request = Request(copy.deepcopy(current.request.environ))
        self.response = copy.deepcopy(current.response)
        self.session = copy.deepcopy(current.session)
        self.T = current.T

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass


class ClientResponse(object):
    def __init__(self, raw, status, headers):
        self.raw = raw
        self.status = status
        self.headers = headers
        self._close = lambda: None
        self._context = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    @cachedprop
    def context(self):
        return ClientContext()

    @cachedprop
    def data(self):
        self._ensure_sequence()
        rv = b''.join(self.iter_encoded())
        return to_native(rv)

    @property
    def is_sequence(self):
        return isinstance(self.raw, (tuple, list))

    def _ensure_sequence(self, mutable=False):
        if self.is_sequence:
            # if we need a mutable object, we ensure it's a list.
            if mutable and not isinstance(self.response, list):
                self.response = list(self.response)
            return
        self.make_sequence()

    def make_sequence(self):
        if not self.is_sequence:
            close = getattr(self.raw, 'close', None)
            self.raw = list(self.iter_encoded())
            self._close = close

    def close(self):
        if hasattr(self.raw, 'close'):
            self.response.close()
        self._close()

    def iter_encoded(self, charset='utf-8'):
        for item in self.raw:
            if isinstance(item, text_type):
                yield item.encode(charset)
            else:
                yield item


class WeppyTestClient(object):
    """This class allows to send requests to a wrapped application."""

    def __init__(self, application, response_wrapper=ClientResponse,
                 use_cookies=True, allow_subdomain_redirects=False):
        self.application = application
        self.response_wrapper = response_wrapper
        if use_cookies:
            self.cookie_jar = TestCookieJar()
        else:
            self.cookie_jar = None
        self.allow_subdomain_redirects = allow_subdomain_redirects

    # def set_cookie(self, server_name, key, value='', max_age=None,
    #                expires=None, path='/', domain=None, secure=None,
    #                httponly=False, charset='utf-8'):
    #     """Sets a cookie in the client's cookie jar.  The server name
    #     is required and has to match the one that is also passed to
    #     the open call.
    #     """
    #     assert self.cookie_jar is not None, 'cookies disabled'
    #     header = dump_cookie(key, value, max_age, expires, path, domain,
    #                          secure, httponly, charset)
    #     environ = create_environ(path, base_url='http://' + server_name)
    #     headers = [('Set-Cookie', header)]
    #     self.cookie_jar.extract_wsgi(environ, headers)

    # def delete_cookie(self, server_name, key, path='/', domain=None):
    #     """Deletes a cookie in the test client."""
    #     self.set_cookie(server_name, key, expires=0, max_age=0,
    #                     path=path, domain=domain)

    def run_wsgi_app(self, environ, buffered=False):
        """Runs the wrapped WSGI app with the given environment."""
        if self.cookie_jar is not None:
            self.cookie_jar.inject_wsgi(environ)
        rv = run_wsgi_app(self.application, environ, buffered=buffered)
        if self.cookie_jar is not None:
            self.cookie_jar.extract_wsgi(environ, rv[2])
        return rv

    def resolve_redirect(self, response, new_loc, environ, buffered=False):
        """Resolves a single redirect and triggers the request again
        directly on this redirect client.
        """
        scheme, netloc, script_root, qs, anchor = url_parse(new_loc)
        base_url = url_unparse((scheme, netloc, '', '', '')).rstrip('/') + '/'

        cur_name = netloc.split(':', 1)[0].split('.')
        real_name = get_host(environ).rsplit(':', 1)[0].split('.')

        if len(cur_name) == 1 and not cur_name[0]:
            allowed = True
        else:
            if self.allow_subdomain_redirects:
                allowed = cur_name[-len(real_name):] == real_name
            else:
                allowed = cur_name == real_name

        if not allowed:
            raise RuntimeError('%r does not support redirect to '
                               'external targets' % self.__class__)

        status_code = int(response[1].split(None, 1)[0])
        if status_code == 307:
            method = environ['REQUEST_METHOD']
        else:
            method = 'GET'

        # For redirect handling we temporarily disable the response
        # wrapper.  This is not threadsafe but not a real concern
        # since the test client must not be shared anyways.
        old_response_wrapper = self.response_wrapper
        self.response_wrapper = None
        try:
            return self.open(path=script_root, base_url=base_url,
                             query_string=qs, as_tuple=True,
                             buffered=buffered, method=method)
        finally:
            self.response_wrapper = old_response_wrapper

    def open(self, *args, **kwargs):
        """Takes the same arguments as the :class:`EnvironBuilder` class with
        some additions:  You can provide a :class:`EnvironBuilder` or a WSGI
        environment as only argument instead of the :class:`EnvironBuilder`
        arguments and two optional keyword arguments (`as_tuple`, `buffered`)
        that change the type of the return value or the way the application is
        executed.

        Additional parameters:

        :param as_tuple: Returns a tuple in the form ``(environ, result)``
        :param buffered: Set this to True to buffer the application run.
                         This will automatically close the application for
                         you as well.
        :param follow_redirects: Set this to True if the `Client` should
                                 follow HTTP redirects.
        """
        as_tuple = kwargs.pop('as_tuple', False)
        buffered = kwargs.pop('buffered', False)
        follow_redirects = kwargs.pop('follow_redirects', False)
        environ = None
        if not kwargs and len(args) == 1:
            if isinstance(args[0], EnvironBuilder):
                environ = args[0].get_environ()
            elif isinstance(args[0], dict):
                environ = args[0]
        if environ is None:
            builder = EnvironBuilder(*args, **kwargs)
            try:
                environ = builder.get_environ()
            finally:
                builder.close()

        response = self.run_wsgi_app(environ, buffered=buffered)

        # handle redirects
        redirect_chain = []
        while 1:
            status_code = int(response[1].split(None, 1)[0])
            if status_code not in (301, 302, 303, 305, 307) \
               or not follow_redirects:
                break
            new_location = response[2]['location']
            if new_location.startswith('/'):
                new_location = \
                    environ['wsgi.url_scheme'] + "://" + environ['HTTP_HOST']
            new_redirect_entry = (new_location, status_code)
            if new_redirect_entry in redirect_chain:
                raise Exception('loop detected')
            redirect_chain.append(new_redirect_entry)
            environ, response = self.resolve_redirect(response, new_location,
                                                      environ,
                                                      buffered=buffered)

        if self.response_wrapper is not None:
            response = self.response_wrapper(*response)
        if as_tuple:
            return environ, response
        return response

    def get(self, *args, **kw):
        """Like open but method is enforced to GET."""
        kw['method'] = 'GET'
        return self.open(*args, **kw)

    def patch(self, *args, **kw):
        """Like open but method is enforced to PATCH."""
        kw['method'] = 'PATCH'
        return self.open(*args, **kw)

    def post(self, *args, **kw):
        """Like open but method is enforced to POST."""
        kw['method'] = 'POST'
        return self.open(*args, **kw)

    def head(self, *args, **kw):
        """Like open but method is enforced to HEAD."""
        kw['method'] = 'HEAD'
        return self.open(*args, **kw)

    def put(self, *args, **kw):
        """Like open but method is enforced to PUT."""
        kw['method'] = 'PUT'
        return self.open(*args, **kw)

    def delete(self, *args, **kw):
        """Like open but method is enforced to DELETE."""
        kw['method'] = 'DELETE'
        return self.open(*args, **kw)

    def options(self, *args, **kw):
        """Like open but method is enforced to OPTIONS."""
        kw['method'] = 'OPTIONS'
        return self.open(*args, **kw)

    def trace(self, *args, **kw):
        """Like open but method is enforced to TRACE."""
        kw['method'] = 'TRACE'
        return self.open(*args, **kw)

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__,
            self.application
        )


def create_environ(*args, **kwargs):
    builder = EnvironBuilder(*args, **kwargs)
    try:
        return builder.get_environ()
    finally:
        builder.close()


def run_wsgi_app(app, environ, buffered=False):
    response = []
    buffer = []

    def start_response(status, headers, exc_info=None):
        if exc_info is not None:
            reraise(*exc_info)
        response[:] = [status, headers]
        return buffer.append

    app_rv = app(environ, start_response)
    close_func = getattr(app_rv, 'close', None)
    app_iter = iter(app_rv)

    # when buffering we emit the close call early and convert the
    # application iterator into a regular list
    if buffered:
        try:
            app_iter = list(app_iter)
        finally:
            if close_func is not None:
                close_func()

    # otherwise we iterate the application iter until we have a response, chain
    # the already received data with the already collected data and wrap it in
    # a new `ClosingIterator` if we need to restore a `close` callable from the
    # original return value.
    else:
        while not response:
            buffer.append(next(app_iter))
        if buffer:
            app_iter = chain(buffer, app_iter)
        if close_func is not None and app_iter is not app_rv:
            app_iter = ClosingIterator(app_iter, close_func)

    return app_iter, response[0], Headers(response[1])



# -*- coding: utf-8 -*-
"""
    weppy.extensions
    ----------------

    Provides base classes to create extensions.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ._compat import iteritems


class Extension(object):
    namespace = None
    default_config = {}

    def __init__(self, app, env, config):
        self.app = app
        self.env = env
        self.config = config
        self.__init_config()

    def __init_config(self):
        for key, dval in iteritems(self.default_config):
            self.config[key] = self.config.get(key, dval)

    def on_load(self):
        pass


class TemplateExtension(object):
    namespace = None
    file_extension = None
    lexers = {}

    def __init__(self, env, config):
        self.env = env
        self.config = config

    def preload(self, path, name):
        return path, name

    def preprocess(self, source, name):
        return source

    def inject(self, context):
        pass


class TemplateLexer(object):
    evaluate_value = True

    def __init__(self, extension):
        self.ext = extension

    def __call__(self, parser, value=None):
        self.parser = parser
        if self.evaluate_value and value is not None:
            value = eval(value, self.parser.context)
        self.process(value)

    @property
    def stack(self):
        return self.parser.stack

    @property
    def top(self):
        return self.parser.stack[-1]

    def process(self, value):
        return value



# -*- coding: utf-8 -*-
"""
    weppy.tools.decorators
    ----------------------

    Provides requires and service decorators.

    :copyright: (c) 2014 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from .._compat import basestring
from ..expose import Expose


class Decorator(object):
    def handler(self):
        pass

    def __call__(self, func):
        obj = Expose.exposing()
        obj.handlers.append(self.handler())
        return func


class requires(Decorator):
    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError(
                'requires usage: @requires(condition, otherwise)'
            )
        if not callable(otherwise) and not isinstance(otherwise, basestring):
            raise SyntaxError(
                "requires 'otherwise' param must be string or callable"
            )
        self.condition = condition
        self.otherwise = otherwise

    def handler(self):
        from ..handlers import RequireHandler
        return RequireHandler(self.condition, self.otherwise)


class service(Decorator):
    def __init__(self, procedure):
        self.procedure = procedure

    @staticmethod
    def json(f):
        return service('json')(f)

    @staticmethod
    def xml(f):
        return service('xml')(f)

    def handler(self):
        from .service import ServiceHandler
        return ServiceHandler(self.procedure)



# -*- coding: utf-8 -*-
"""
    weppy.tools.mail
    ----------------

    provide a mailing system

    adapted from the original code of web2py (http://www.web2py.com)

    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>
    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import os
import time
import smtplib

from .._compat import iteritems, PY2
from ..utils import read_file
from ..datastructures import sdict
from ..libs.contenttype import contenttype

if PY2:
    from email.MIMEBase import MIMEBase
    from email import MIMEMultipart, MIMEText, Encoders, Header, \
        message_from_string, Charset
else:
    from email import message_from_string
    from email.mime.base import MIMEBase
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
    from email import encoders as Encoders
    from email.header import Header
    from email.charset import Charset


class Mail(object):
    """
    Class for configuring and sending emails with alternative text / html
    body, multiple attachments and encryption support

    Works with SMTP and Google App Engine.
    """

    class Attachment(MIMEBase):
        """
        Email attachment

        Arguments:

            payload: path to file or file-like object with read() method
            filename: name of the attachment stored in message; if set to
                      None, it will be fetched from payload path; file-like
                      object payload must have explicit filename specified
            content_id: id of the attachment; automatically contained within
                        < and >
            content_type: content type of the attachment; if set to None,
                          it will be fetched from filename using gluon.contenttype
                          module
            encoding: encoding of all strings passed to this function (except
                      attachment body)

        Content ID is used to identify attachments within the html body;
        in example, attached image with content ID 'photo' may be used in
        html message as a source of img tag <img src="cid:photo" />.

        Examples:

            #Create attachment from text file:
            attachment = Mail.Attachment('/path/to/file.txt')

            Content-Type: text/plain
            MIME-Version: 1.0
            Content-Disposition: attachment; filename="file.txt"
            Content-Transfer-Encoding: base64

            SOMEBASE64CONTENT=

            #Create attachment from image file with custom filename and cid:
            attachment = Mail.Attachment('/path/to/file.png',
                                             filename='photo.png',
                                             content_id='photo')

            Content-Type: image/png
            MIME-Version: 1.0
            Content-Disposition: attachment; filename="photo.png"
            Content-Id: <photo>
            Content-Transfer-Encoding: base64

            SOMEOTHERBASE64CONTENT=
        """

        def __init__(
            self,
            payload,
            filename=None,
            content_id=None,
            content_type=None,
                encoding='utf-8'):
            if isinstance(payload, str):
                if filename is None:
                    filename = os.path.basename(payload)
                payload = read_file(payload, 'rb')
            else:
                if filename is None:
                    raise Exception('Missing attachment name')
                payload = payload.read()
            filename = filename.encode(encoding)
            if content_type is None:
                content_type = contenttype(filename)
            self.my_filename = filename
            self.my_payload = payload
            MIMEBase.__init__(self, *content_type.split('/', 1))
            self.set_payload(payload)
            self['Content-Disposition'] = 'attachment; filename="%s"' % filename
            if content_id is not None:
                self['Content-Id'] = '<%s>' % content_id.encode(encoding)
            Encoders.encode_base64(self)

    def __init__(self, app, server=None, sender=None, login=None, tls=True):
        """
        Main Mail object

        Arguments:

            server: SMTP server address in address:port notation
            sender: sender email address
            login: sender login name and password in login:password notation
                   or None if no authentication is required
            tls: enables/disables encryption (True by default)

        In Google App Engine use:

            server='gae'

        For sake of backward compatibility all fields are optional and default
        to None, however, to be able to send emails at least server and sender
        must be specified. They are available under following fields:

            mail.settings.server
            mail.settings.sender
            mail.settings.login

        When server is 'logging', email is logged but not sent (debug mode)

        Optionally you can use PGP encryption or X509:

            mail.settings.cipher_type = None
            mail.settings.gpg_home = None
            mail.settings.sign = True
            mail.settings.sign_passphrase = None
            mail.settings.encrypt = True
            mail.settings.x509_sign_keyfile = None
            mail.settings.x509_sign_certfile = None
            mail.settings.x509_nocerts = False
            mail.settings.x509_crypt_certfiles = None

            cipher_type       : None
                                gpg - need a python-pyme package and gpgme lib
                                x509 - smime
            gpg_home          : you can set a GNUPGHOME environment variable
                                to specify home of gnupg
            sign              : sign the message (True or False)
            sign_passphrase   : passphrase for key signing
            encrypt           : encrypt the message
                             ... x509 only ...
            x509_sign_keyfile : the signers private key filename (PEM format)
            x509_sign_certfile: the signers certificate filename (PEM format)
            x509_nocerts      : if True then no attached certificate in mail
            x509_crypt_certfiles: the certificates file to encrypt the messages
                                  with can be a file name or a list of
                                  file names (PEM format)

        Examples:

            #Create Mail object with authentication data for remote server:
            mail = Mail('example.com:25', 'me@example.com', 'me:password')
        """

        self.app = app
        # read config from app

        settings = self.settings = sdict()
        settings.server = server
        settings.sender = sender
        settings.login = login
        settings.tls = tls
        settings.hostname = None
        settings.ssl = False
        settings.cipher_type = None
        settings.gpg_home = None
        settings.sign = True
        settings.sign_passphrase = None
        settings.encrypt = True
        settings.x509_sign_keyfile = None
        settings.x509_sign_certfile = None
        settings.x509_nocerts = False
        settings.x509_crypt_certfiles = None
        settings.debug = False
        settings.lock_keys = True
        self.result = {}
        self.error = None

    def send(
        self,
        to,
        subject='[no subject]',
        message='[no message]',
        attachments=None,
        cc=None,
        bcc=None,
        reply_to=None,
        sender=None,
        encoding='utf-8',
        raw=False,
        headers={}
    ):
        """
        Sends an email using data specified in constructor

        Arguments:

            to: list or tuple of receiver addresses; will also accept single
                object
            subject: subject of the email
            message: email body text; depends on type of passed object:
                     if 2-list or 2-tuple is passed: first element will be
                     source of plain text while second of html text;
                     otherwise: object will be the only source of plain text
                     and html source will be set to None;
                     If text or html source is:
                     None: content part will be ignored,
                     string: content part will be set to it,
                     file-like object: content part will be fetched from
                                       it using it's read() method
            attachments: list or tuple of Mail.Attachment objects; will also
                         accept single object
            cc: list or tuple of carbon copy receiver addresses; will also
                accept single object
            bcc: list or tuple of blind carbon copy receiver addresses; will
                also accept single object
            reply_to: address to which reply should be composed
            encoding: encoding of all strings passed to this method (including
                      message bodies)
            headers: dictionary of headers to refine the headers just before
                     sending mail, e.g. {'Return-Path' : 'bounces@example.org'}

        Examples:

            #Send plain text message to single address:
            mail.send('you@example.com',
                      'Message subject',
                      'Plain text body of the message')

            #Send html message to single address:
            mail.send('you@example.com',
                      'Message subject',
                      '<html>Plain text body of the message</html>')

            #Send text and html message to three addresses (two in cc):
            mail.send('you@example.com',
                      'Message subject',
                      ('Plain text body', '<html>html body</html>'),
                      cc=['other1@example.com', 'other2@example.com'])

            #Send html only message with image attachment available from
            the message by 'photo' content id:
            mail.send('you@example.com',
                      'Message subject',
                      (None, '<html><img src="cid:photo" /></html>'),
                      Mail.Attachment('/path/to/photo.jpg'
                                      content_id='photo'))

            #Send email with two attachments and no body text
            mail.send('you@example.com,
                      'Message subject',
                      None,
                      [Mail.Attachment('/path/to/fist.file'),
                       Mail.Attachment('/path/to/second.file')])

        Returns True on success, False on failure.

        Before return, method updates two object's fields:
        self.result: return value of smtplib.SMTP.sendmail() or GAE's
                     mail.send_mail() method
        self.error: Exception message or None if above was successful
        """

        # We don't want to use base64 encoding for unicode mail
        Charset.add_charset('utf-8', Charset.QP, Charset.QP, 'utf-8')

        def encode_header(key):
            if [c for c in key if 32 > ord(c) or ord(c) > 127]:
                return Header.Header(key.encode('utf-8'), 'utf-8')
            else:
                return key

        # encoded or raw text
        def encoded_or_raw(text):
            if raw:
                text = encode_header(text)
            return text

        sender = sender or self.settings.sender

        if not isinstance(self.settings.server, str):
            raise Exception('Server address not specified')
        if not isinstance(sender, str):
            raise Exception('Sender address not specified')

        if not raw and attachments:
            # Use multipart/mixed if there is attachments
            payload_in = MIMEMultipart.MIMEMultipart('mixed')
        elif raw:
            # no encoding configuration for raw messages
            if not isinstance(message, basestring):
                message = message.read()
            if isinstance(message, unicode):
                text = message.encode('utf-8')
            elif not encoding == 'utf-8':
                text = message.decode(encoding).encode('utf-8')
            else:
                text = message
            # No charset passed to avoid transport encoding
            # NOTE: some unicode encoded strings will produce
            # unreadable mail contents.
            payload_in = MIMEText.MIMEText(text)
        if to:
            if not isinstance(to, (list, tuple)):
                to = [to]
        else:
            raise Exception('Target receiver address not specified')
        if cc:
            if not isinstance(cc, (list, tuple)):
                cc = [cc]
        if bcc:
            if not isinstance(bcc, (list, tuple)):
                bcc = [bcc]
        if message is None:
            text = html = None
        elif isinstance(message, (list, tuple)):
            text, html = message
        elif message.strip().startswith('<html') and \
                message.strip().endswith('</html>'):
            text = self.settings.server == 'gae' and message or None
            html = message
        else:
            text = message
            html = None

        if (text is not None or html is not None) and (not raw):

            if text is not None:
                if not isinstance(text, basestring):
                    text = text.read()
                if isinstance(text, unicode):
                    text = text.encode('utf-8')
                elif not encoding == 'utf-8':
                    text = text.decode(encoding).encode('utf-8')
            if html is not None:
                if not isinstance(html, basestring):
                    html = html.read()
                if isinstance(html, unicode):
                    html = html.encode('utf-8')
                elif not encoding == 'utf-8':
                    html = html.decode(encoding).encode('utf-8')

            # Construct mime part only if needed
            if text is not None and html:
                # We have text and html we need multipart/alternative
                attachment = MIMEMultipart.MIMEMultipart('alternative')
                attachment.attach(MIMEText.MIMEText(text, _charset='utf-8'))
                attachment.attach(
                    MIMEText.MIMEText(html, 'html', _charset='utf-8'))
            elif text is not None:
                attachment = MIMEText.MIMEText(text, _charset='utf-8')
            elif html:
                attachment = \
                    MIMEText.MIMEText(html, 'html', _charset='utf-8')

            if attachments:
                # If there is attachments put text and html into
                # multipart/mixed
                payload_in.attach(attachment)
            else:
                # No attachments no multipart/mixed
                payload_in = attachment

        if (attachments is None) or raw:
            pass
        elif isinstance(attachments, (list, tuple)):
            for attachment in attachments:
                payload_in.attach(attachment)
        else:
            payload_in.attach(attachments)

        #######################################################
        #                      CIPHER                         #
        #######################################################
        cipher_type = self.settings.cipher_type
        sign = self.settings.sign
        sign_passphrase = self.settings.sign_passphrase
        encrypt = self.settings.encrypt
        #######################################################
        #                       GPGME                         #
        #######################################################
        if cipher_type == 'gpg':
            if self.settings.gpg_home:
                # Set GNUPGHOME environment variable to set home of gnupg
                import os
                os.environ['GNUPGHOME'] = self.settings.gpg_home
            if not sign and not encrypt:
                self.error = "No sign and no encrypt is set but cipher type to gpg"
                return False

            # need a python-pyme package and gpgme lib
            from pyme import core, errors
            from pyme.constants.sig import mode
            ############################################
            #                   sign                   #
            ############################################
            if sign:
                import string
                core.check_version(None)
                pin = string.replace(payload_in.as_string(), '\n', '\r\n')
                plain = core.Data(pin)
                sig = core.Data()
                c = core.Context()
                c.set_armor(1)
                c.signers_clear()
                # search for signing key for From:
                for sigkey in c.op_keylist_all(sender, 1):
                    if sigkey.can_sign:
                        c.signers_add(sigkey)
                if not c.signers_enum(0):
                    self.error = 'No key for signing [%s]' % sender
                    return False
                c.set_passphrase_cb(lambda x, y, z: sign_passphrase)
                try:
                    # make a signature
                    c.op_sign(plain, sig, mode.DETACH)
                    sig.seek(0, 0)
                    # make it part of the email
                    payload = MIMEMultipart.MIMEMultipart('signed',
                                                          boundary=None,
                                                          _subparts=None,
                                                          **dict(
                                                          micalg="pgp-sha1",  # NOQA
                                                          protocol="application/pgp-signature"))  # NOQA
                    # insert the origin payload
                    payload.attach(payload_in)
                    # insert the detached signature
                    p = MIMEBase("application", 'pgp-signature')
                    p.set_payload(sig.read())
                    payload.attach(p)
                    # it's just a trick to handle the no encryption case
                    payload_in = payload
                except errors.GPGMEError as ex:
                    self.error = "GPG error: %s" % ex.getstring()
                    return False
            ############################################
            #                  encrypt                 #
            ############################################
            if encrypt:
                core.check_version(None)
                plain = core.Data(payload_in.as_string())
                cipher = core.Data()
                c = core.Context()
                c.set_armor(1)
                # collect the public keys for encryption
                recipients = []
                rec = to[:]
                if cc:
                    rec.extend(cc)
                if bcc:
                    rec.extend(bcc)
                for addr in rec:
                    c.op_keylist_start(addr, 0)
                    r = c.op_keylist_next()
                    if r is None:
                        self.error = 'No key for [%s]' % addr
                        return False
                    recipients.append(r)
                try:
                    # make the encryption
                    c.op_encrypt(recipients, 1, plain, cipher)
                    cipher.seek(0, 0)
                    # make it a part of the email
                    payload = MIMEMultipart.MIMEMultipart('encrypted',
                                                          boundary=None,
                                                          _subparts=None,
                                                          **dict(protocol="application/pgp-encrypted"))
                    p = MIMEBase("application", 'pgp-encrypted')
                    p.set_payload("Version: 1\r\n")
                    payload.attach(p)
                    p = MIMEBase("application", 'octet-stream')
                    p.set_payload(cipher.read())
                    payload.attach(p)
                except errors.GPGMEError as ex:
                    self.error = "GPG error: %s" % ex.getstring()
                    return False
        #######################################################
        #                       X.509                         #
        #######################################################
        elif cipher_type == 'x509':
            if not sign and not encrypt:
                self.error = "No sign and no encrypt is set but cipher type to x509"
                return False
            x509_sign_keyfile = self.settings.x509_sign_keyfile
            if self.settings.x509_sign_certfile:
                x509_sign_certfile = self.settings.x509_sign_certfile
            else:
                # if there is no sign certfile we'll assume the
                # cert is in keyfile
                x509_sign_certfile = self.settings.x509_sign_keyfile
            # crypt certfiles could be a string or a list
            x509_crypt_certfiles = self.settings.x509_crypt_certfiles
            x509_nocerts = self.settings.x509_nocerts

            # need m2crypto
            try:
                from M2Crypto import BIO, SMIME, X509
            except Exception as e:
                self.error = "Can't load M2Crypto module"
                return False
            msg_bio = BIO.MemoryBuffer(payload_in.as_string())
            s = SMIME.SMIME()

            #                   SIGN
            if sign:
                #key for signing
                try:
                    s.load_key(x509_sign_keyfile, x509_sign_certfile,
                               callback=lambda x: sign_passphrase)
                except Exception as e:
                    self.error = "Something went wrong on certificate / private key loading: <%s>" % str(e)
                    return False
                try:
                    if x509_nocerts:
                        flags = SMIME.PKCS7_NOCERTS
                    else:
                        flags = 0
                    if not encrypt:
                        flags += SMIME.PKCS7_DETACHED
                    p7 = s.sign(msg_bio, flags=flags)
                    msg_bio = BIO.MemoryBuffer(payload_in.as_string(
                    ))  # Recreate coz sign() has consumed it.
                except Exception as e:
                    self.error = "Something went wrong on signing: <%s> %s" % (
                        str(e), str(flags))
                    return False

            #                   ENCRYPT
            if encrypt:
                try:
                    sk = X509.X509_Stack()
                    if not isinstance(x509_crypt_certfiles, (list, tuple)):
                        x509_crypt_certfiles = [x509_crypt_certfiles]

                    # make an encryption cert's stack
                    for x in x509_crypt_certfiles:
                        sk.push(X509.load_cert(x))
                    s.set_x509_stack(sk)

                    s.set_cipher(SMIME.Cipher('des_ede3_cbc'))
                    tmp_bio = BIO.MemoryBuffer()
                    if sign:
                        s.write(tmp_bio, p7)
                    else:
                        tmp_bio.write(payload_in.as_string())
                    p7 = s.encrypt(tmp_bio)
                except Exception as e:
                    self.error = "Something went wrong on encrypting: <%s>" % str(e)
                    return False

            #                 Final stage in sign and encryption
            out = BIO.MemoryBuffer()
            if encrypt:
                s.write(out, p7)
            else:
                if sign:
                    s.write(out, p7, msg_bio, SMIME.PKCS7_DETACHED)
                else:
                    out.write('\r\n')
                    out.write(payload_in.as_string())
            out.close()
            st = str(out.read())
            payload = message_from_string(st)
        else:
            # no cryptography process as usual
            payload = payload_in

        payload['From'] = encoded_or_raw(sender.decode(encoding))
        origTo = to[:]
        if to:
            payload['To'] = encoded_or_raw(', '.join(to).decode(encoding))
        if reply_to:
            payload['Reply-To'] = encoded_or_raw(reply_to.decode(encoding))
        if cc:
            payload['Cc'] = encoded_or_raw(', '.join(cc).decode(encoding))
            to.extend(cc)
        if bcc:
            to.extend(bcc)
        payload['Subject'] = encoded_or_raw(subject.decode(encoding))
        payload['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S +0000",
                                        time.gmtime())
        for k, v in iteritems(headers):
            payload[k] = encoded_or_raw(v.decode(encoding))
        result = {}
        try:
            if self.settings.server == 'logging':
                self.app.log.warn('email not sent\n%s\nFrom: %s\nTo: %s\nSubject: %s\n\n%s\n%s\n' %
                            ('-' * 40, sender,  # NOQA
                             ', '.join(to), subject,
                             text or html, '-' * 40))
            elif self.settings.server == 'gae':
                xcc = dict()
                if cc:
                    xcc['cc'] = cc
                if bcc:
                    xcc['bcc'] = bcc
                if reply_to:
                    xcc['reply_to'] = reply_to
                from google.appengine.api import mail
                attachments = attachments and [(a.my_filename, a.my_payload) for a in attachments if not raw]
                if attachments:
                    result = mail.send_mail(
                        sender=sender, to=origTo,
                        subject=subject, body=text, html=html,
                        attachments=attachments, **xcc)
                elif html and (not raw):
                    result = mail.send_mail(
                        sender=sender, to=origTo,
                        subject=subject, body=text, html=html, **xcc)
                else:
                    result = mail.send_mail(
                        sender=sender, to=origTo,
                        subject=subject, body=text, **xcc)
            else:
                smtp_args = self.settings.server.split(':')
                if self.settings.ssl:
                    server = smtplib.SMTP_SSL(*smtp_args)
                else:
                    server = smtplib.SMTP(*smtp_args)
                if self.settings.tls and not self.settings.ssl:
                    server.ehlo(self.settings.hostname)
                    server.starttls()
                    server.ehlo(self.settings.hostname)
                if self.settings.login:
                    server.login(*self.settings.login.split(':', 1))
                result = server.sendmail(
                    sender, to, payload.as_string())
                server.quit()
        except Exception as e:
            self.app.log.warn('Mail.send failure:%s' % e)
            self.result = result
            self.error = e
            return False
        self.result = result
        self.error = None
        return True



# -*- coding: utf-8 -*-
"""
    tests.session
    ----------------

    Test weppy session module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import pytest

from weppy.testing.env import EnvironBuilder
from weppy.sessions import SessionCookieManager


@pytest.fixture(scope='module')
def current():
    from weppy.globals import current
    builder = EnvironBuilder()
    current.initialize(builder.get_environ())
    return current


def test_session_cookie(current):
    session_cookie = SessionCookieManager(
        key='sid',
        secure=True,
        domain='localhost'
    )
    assert session_cookie.key == 'sid'
    assert session_cookie.secure is True
    assert session_cookie.domain == 'localhost'

    session_cookie.on_start()
    assert current.session._expiration == 3600

    session_cookie.on_end()
    cookie = str(current.response.cookies)
    assert 'wpp_session_data_' in cookie
    assert 'Domain=localhost;' in cookie
    # Python 3.5 make Secure attribute in cookie capitalized
    assert 'secure' in cookie.lower()

    current.request.cookies = current.response.cookies
    session_cookie.on_start()
    assert current.session._expiration == 3600



# -*- coding: utf-8 -*-
"""
    weppy.app
    ---------

    Provides the central application object.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import sys
import os
import click
from yaml import load as ymlload
from ._compat import basestring
from ._internal import get_root_path, create_missing_app_folders
from .datastructures import sdict, ConfigData
from .expose import Expose, url
from .extensions import Extension, TemplateExtension
from .globals import current
from .templating.core import Templater
from .utils import dict_to_sdict, cachedprop, read_file
from .wsgi import error_handler


class App(object):
    debug = None
    test_client_class = None

    def __init__(self, import_name, root_path=None,
                 template_folder='templates', config_folder='config'):
        self.import_name = import_name
        #: Set paths for the application
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        self.static_path = os.path.join(self.root_path, "static")
        self.template_path = os.path.join(self.root_path, template_folder)
        self.config_path = os.path.join(self.root_path, config_folder)
        #: The click command line context for this application.
        self.cli = click.Group(self)
        #: Init the configuration
        self.config = ConfigData()
        self.config.hostname_default = None
        self.config.static_version = None
        self.config.static_version_urls = None
        self.config.url_default_namespace = None
        self.config.templates_auto_reload = False
        #: Trying to create needed folders
        create_missing_app_folders(self)
        #: init expose module
        Expose.application = self
        self.error_handlers = {}
        self.template_default_extension = '.html'
        #: init logger
        self._logger = None
        self.logger_name = self.import_name
        #: set request.now reference
        self.now_reference = "utc"
        #: init languages
        self.languages = []
        self.language_default = None
        self.language_force_on_url = False
        self.language_write = False
        #: init extensions
        self.ext = sdict()
        self._extensions_env = sdict()
        self.template_extensions = []
        self.template_preloaders = {}
        self.template_lexers = {}
        self.templater = Templater(self)

    @cachedprop
    def name(self):
        """The name of the application. This is usually the import name
        with the difference that it's guessed from the run file if the
        import name is main.
        """
        if self.import_name == '__main__':
            fn = getattr(sys.modules['__main__'], '__file__', None)
            if fn is None:
                rv = '__main__'
            else:
                rv = os.path.splitext(os.path.basename(fn))[0]
        else:
            rv = self.import_name
        return rv

    @property
    def route(self):
        return Expose

    @property
    def common_handlers(self):
        return self.route.common_handlers

    @common_handlers.setter
    def common_handlers(self, handlers):
        self.route.common_handlers = handlers

    @property
    def common_helpers(self):
        return self.route.common_helpers

    @common_helpers.setter
    def common_helpers(self, helpers):
        self.route.common_helpers = helpers

    @property
    def routing_processors(self):
        return self.route.processors

    def on_error(self, code):
        def decorator(f):
            self.error_handlers[code] = f
            return f
        return decorator

    @property
    def command(self):
        return self.cli.command

    @property
    def log(self):
        if self._logger and self._logger.name == self.logger_name:
            return self._logger
        from .logger import _logger_lock, create_logger
        with _logger_lock:
            if self._logger and self._logger.name == self.logger_name:
                return self._logger
            self._logger = rv = create_logger(self)
            return rv

    def render_template(self, filename):
        ctx = {'current': current, 'url': url}
        return self.templater.render(self.template_path, filename, ctx)

    def config_from_yaml(self, filename, namespace=None):
        #: import configuration from yaml files
        rc = read_file(os.path.join(self.config_path, filename))
        rc = ymlload(rc)
        c = self.config if namespace is None else self.config[namespace]
        for key, val in rc.items():
            c[key] = dict_to_sdict(val)

    #: Creates the extensions' environments and configs
    def __init_extension(self, ext):
        if ext.namespace is None:
            ext.namespace = ext.__name__
        if self._extensions_env[ext.namespace] is None:
            self._extensions_env[ext.namespace] = sdict()
        return self._extensions_env[ext.namespace], self.config[ext.namespace]

    #: Add an extension to application
    def use_extension(self, ext):
        if not issubclass(ext, Extension):
            raise RuntimeError('%s is an invalid weppy extension' %
                               ext.__name__)
        ext_env, ext_config = self.__init_extension(ext)
        self.ext[ext.__name__] = ext(self, ext_env, ext_config)
        self.ext[ext.__name__].on_load()

    #: Add a template extension to application
    def add_template_extension(self, ext):
        if not issubclass(ext, TemplateExtension):
            raise RuntimeError('%s is an invalid weppy template extension' %
                               ext.__name__)
        ext_env, ext_config = self.__init_extension(ext)
        self.template_extensions.append(ext(ext_env, ext_config))
        fext = self.template_extensions[-1].file_extension
        if fext is not None and isinstance(fext, basestring):
            if fext not in self.template_preloaders.keys():
                self.template_preloaders[fext] = []
            self.template_preloaders[fext].append(self.template_extensions[-1])
        lexers = self.template_extensions[-1].lexers
        for name, lexer in lexers.items():
            self.template_lexers[name] = lexer(self.template_extensions[-1])

    def make_shell_context(self, context={}):
        """Returns the shell context for an interactive shell for this
        application.  This runs all the registered shell context
        processors.
        """
        context['app'] = self
        return context

    def _run(self, host, port):
        from .libs.rocket import Rocket
        r = Rocket((host, port), 'wsgi', {'wsgi_app': self})
        r.start()

    def run(self, host=None, port=None, reloader=True, debug=True):
        if host is None:
            host = "127.0.0.1"
        if port is None:
            port = 8000
        self.debug = debug
        if os.environ.get('WEPPY_RUN_MAIN') != 'true':
            quit_msg = "(press CTRL+C to quit)"
            self.log.info("> weppy application %s running on http://%s:%i %s" %
                          (self.import_name, host, port, quit_msg))
        if reloader:
            from ._reloader import run_with_reloader
            run_with_reloader(self, host, port)
        else:
            self._run(host, port)

    def test_client(self, use_cookies=True, **kwargs):
        tclass = self.test_client_class
        if tclass is None:
            from .testing import WeppyTestClient
            tclass = WeppyTestClient
        return tclass(self, use_cookies=use_cookies, **kwargs)

    def wsgi_handler(self, environ, start_request):
        return error_handler(self, environ, start_request)

    def __call__(self, environ, start_request):
        return self.wsgi_handler(environ, start_request)


class AppModule(object):
    def __init__(self, app, name, import_name, template_folder=None,
                 template_path=None, url_prefix=None, hostname=None,
                 root_path=None):
        self.app = app
        self.name = name
        self.import_name = import_name
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        #: template_folder is referred to application template_path
        self.template_folder = template_folder
        #: template_path is referred to module root_directory
        if template_path and not template_path.startswith("/"):
            template_path = self.root_path + template_path
        self.template_path = template_path
        # how to route static?
        # and.. do we want this?? I think not..
        #if static_folder:
        #    self.static_folder = self.root_path+"/"+static_folder
        #if static_prefix:
        #    self.static_folder = self.app.static_folder+"/"+static_prefix
        self.url_prefix = url_prefix
        self.hostname = hostname
        self.common_handlers = []
        self.common_helpers = []

    def route(self, path=None, name=None, template=None, **kwargs):
        if name is not None and "." in name:
            raise RuntimeError(
                "App modules' route names should not contains dots"
            )
        name = self.name + "." + (name or "")
        handlers = kwargs.get('handlers', [])
        helpers = kwargs.get('helpers', [])
        if self.common_handlers:
            handlers = self.common_handlers + handlers
        kwargs['handlers'] = handlers
        if self.common_helpers:
            helpers = self.common_helpers + helpers
        kwargs['helpers'] = helpers
        return self.app.route(
            path=path, name=name, template=template, prefix=self.url_prefix,
            template_folder=self.template_folder,
            template_path=self.template_path, hostname=self.hostname, **kwargs)



# -*- coding: utf-8 -*-
"""
    tests.globals
    ----------------

    Test weppy globals module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""


from weppy.globals import Request, Response, Current
from weppy.testing.env import EnvironBuilder


def test_request():
    builder = EnvironBuilder(
        path='/?foo=bar',
        method='GET',
    )
    request = Request(builder.get_environ())

    assert request.params == {'foo': 'bar'}
    assert request.client == '127.0.0.1'


def test_response():
    builder = EnvironBuilder(
        path='/?foo=bar',
        method='GET',
    )
    response = Response(builder.get_environ())

    assert response.status == 200
    assert response.headers == {'Content-Type': 'text/html; charset=utf-8'}


def test_current():
    builder = EnvironBuilder(
        path='/?foo=bar',
        method='GET',
    )
    current = Current()
    environ = builder.get_environ()
    current.initialize(environ)

    assert current.environ == environ
    assert isinstance(current.request, Request)
    assert isinstance(current.response, Response)



#!/usr/bin/env python
# -*- coding: utf8 -*-
# Plural-Forms for tr (Turkish)

# Turkish language has ONE form!
nplurals = 1

# Always returns 0:
get_plural_id = lambda n: 0

# Construct and return plural form of *word* using
# *plural_id* (which ALWAYS>0). This function will be executed
# for words (or phrases) not found in plural_dict dictionary
# construct_plural_form = lambda word, plural_id: word



# -*- coding: utf-8 -*-
"""
    tests.expose
    ----------------

    Test weppy expose module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import pytest
from weppy import App
from weppy.http import HTTP
from weppy.expose import url
from weppy.testing.env import EnvironBuilder


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    app.languages = ['en', 'it']
    app.language_default = 'en'
    app.language_force_on_url = True
    return app


def test_expose_valid_route(app):
    from weppy.globals import current
    builder = EnvironBuilder('/it/test_route')
    current.initialize(builder.get_environ())

    @app.route()
    def test_route():
        return 'Test Router'

    response = current.response
    app.route.dispatch()
    assert response.status == 200
    assert response.output == 'Test Router'
    assert current.request.language == 'it'


def test_expose_not_found_route(app):
    from weppy.globals import current
    builder = EnvironBuilder('/')
    current.initialize(builder.get_environ())

    @app.route()
    def test_route():
        return 'Test Router'

    try:
        app.route.dispatch()
    except HTTP as exc:
        assert exc.status_code == 404
        assert exc.body == [b'Invalid action\n']


def test_expose_exception_route(app):
    from weppy.globals import current
    builder = EnvironBuilder('/test_route')
    current.initialize(builder.get_environ())

    @app.route()
    def test_route():
        raise HTTP(404, 'Not found')

    try:
        app.route.dispatch()
    except HTTP as exc:
        assert exc.status_code == 404
        assert exc.body == [b'Not found']


def test_static_url():
    link = url('static', 'file')
    assert link == '/static/file'


def test_module_url(app):
    from weppy.globals import current
    builder = EnvironBuilder('/')
    current.initialize(builder.get_environ())
    current.request.language = 'it'

    @app.route()
    def test_route():
        return 'Test Router'

    link = url('test_route')
    assert link == '/it/test_route'



# -*- coding: utf-8 -*-
"""
    weppy.validators
    ----------------

    Implements validators for pyDAL.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from .._compat import iteritems, basestring
from .basic import Validator, isntEmpty, isEmptyOr, Equals, Matches, \
    hasLength, Not, Allow, isEmpty, Any
from .consist import isInt, isFloat, isDecimal, isDate, isTime, isDatetime, \
    isEmail, isJSON, isUrl, isIP, isImage, isAlphanumeric, isList
from .inside import inRange, inSet, inDB, notInDB
from .process import Cleanup, Crypt, Lower, Urlify, Upper


class ValidateFromDict(object):
    numkeys = {'gt': False, 'gte': True, 'lt': False, 'lte': True}

    def __init__(self):
        self.is_validators = {
            'int': isInt, 'float': isFloat, 'decimal': isDecimal,
            'date': isDate, 'time': isTime, 'datetime': isDatetime,
            'email': isEmail, 'url': isUrl, 'ip': isIP, 'json': isJSON,
            'image': isImage, 'alphanumeric': isAlphanumeric
        }
        self.proc_validators = {
            'crypt': Crypt, 'lower': Lower, 'upper': Upper, 'urlify': Urlify,
            'clean': Cleanup
        }

    def parse_num_comparisons(self, data, minv=None, maxv=None):
        inclusions = [True, False]
        for key, include in iteritems(self.numkeys):
            val = data.get(key)
            if val is not None:
                if key[0] == "g":
                    minv = val
                    inclusions[0] = include
                elif key[0] == "l":
                    maxv = val
                    inclusions[1] = include
        return minv, maxv, inclusions

    def parse_is(self, data):
        #: map types with fields
        if isinstance(data, basestring):
            #: map {'is': 'int'}
            key = data
            options = {}
        elif isinstance(data, dict):
            #: map {'is': {'float': {'dot': ','}}}
            key = list(data)[0]
            options = data[key]
        else:
            raise SyntaxError("'is' validator accepts only string or dict")
        validator = self.is_validators.get(key)
        return validator(**options) if validator else None

    def parse_is_list(self, data):
        #: map types with 'list' fields
        key = ''
        options = {}
        suboptions = {}
        lopts = ['splitter']
        if isinstance(data, basestring):
            #: map {'is': 'list:int'}
            key = data
        elif isinstance(data, dict):
            #: map {'is': {'list:float': {'dot': ',', 'splitter': ';'}}}
            key = list(data)[0]
            suboptions = data[key]
            for opt in list(suboptions):
                if opt in lopts:
                    options[opt] = suboptions[opt]
                    del suboptions[opt]
        else:
            raise SyntaxError("'is' validator accepts only string or dict")
        try:
            keyspecs = key.split(':')
            subkey = keyspecs[1].strip()
            assert keyspecs[0].strip() == 'list'
        except:
            subkey = '_missing_'
        validator = self.is_validators.get(subkey)
        return isList(
            [validator(**suboptions)], **options) if validator else None

    def parse_reference(self, field):
        ref_table = None
        multiple = None
        if field.type.startswith('reference'):
            multiple = False
        elif field.type.startswith('list:reference'):
            multiple = True
        if multiple is not None:
            ref_table = field.type.split(' ')[1]
        return ref_table, multiple

    def __call__(self, field, data):
        validators = []
        #: parse 'presence' and 'empty'
        presence = data.get('presence')
        empty = data.get('empty')
        if presence is None and empty is not None:
            presence = not empty
        #: parse 'is'
        _is = data.get('is')
        if _is is not None:
            validator = self.parse_is(_is) or self.parse_is_list(_is)
            if validator is None:
                raise SyntaxError(
                    "Unknown type %s for 'is' validator" % data
                )
            validators.append(validator)
        #: parse 'len'
        _len = data.get('len')
        if _len is not None:
            if isinstance(_len, int):
                #: allows {'len': 2}
                validators.append(
                    hasLength(_len+1, _len, message='Enter %(min)g characters')
                )
            else:
                #: allows
                #  {'len': {'gt': 1, 'gte': 2, 'lt': 5, 'lte' 6}}
                #  {'len': {'range': (2, 6)}}
                if _len.get('range') is not None:
                    minv, maxv = _len['range']
                    inc = (True, False)
                else:
                    minv, maxv, inc = self.parse_num_comparisons(_len, 0, 256)
                validators.append(hasLength(maxv, minv, inc))
        #: parse 'in'
        _dbset = None
        _in = data.get('in', [])
        if _in:
            if isinstance(_in, (list, tuple, set)):
                #: allows {'in': [1, 2]}
                validators.append(inSet(_in))
            elif isinstance(_in, dict):
                options = {}
                #: allows {'in': {'range': (1, 5)}}
                _range = _in.get('range')
                if isinstance(_range, (tuple, list)):
                    validators.append(inRange(_range[0], _range[1]))
                #: allows {'in': {'set': [1, 5]}} with options
                _set = _in.get('set')
                if isinstance(_set, (list, tuple, set)):
                    opt_keys = [key for key in list(_in) if key != 'set']
                    for key in opt_keys:
                        options[key] = _in[key]
                    validators.append(inSet(_set, **options))
                #: allows {'in': {'dbset': lambda db: db.where(query)}}
                _dbset = _in.get('dbset')
                if callable(_dbset):
                    ref_table, multiple = self.parse_reference(field)
                    if ref_table:
                        opt_keys = [key for key in list(_in) if key != 'dbset']
                        for key in opt_keys:
                            options[key] = _in[key]
                        validators.append(
                            inDB(
                                field.db, ref_table, dbset=_dbset,
                                multiple=multiple, **options))
                    else:
                        raise SyntaxError(
                            "'in:dbset' validator needs a reference field")
            else:
                raise SyntaxError(
                    "'in' validator accepts only a set or a dict")
        #: parse 'gt', 'gte', 'lt', 'lte'
        minv, maxv, inc = self.parse_num_comparisons(data)
        if minv is not None or maxv is not None:
            validators.append(inRange(minv, maxv, inc))
        #: parse 'equals'
        if 'equals' in data:
            validators.append(Equals(data['equals']))
        #: parse 'match'
        if 'match' in data:
            if isinstance(data['match'], dict):
                validators.append(Matches(**data['match']))
            else:
                validators.append(Matches(data['match']))
        #: parse transforming validators
        for key, vclass in iteritems(self.proc_validators):
            if key in data:
                options = {}
                if isinstance(data[key], dict):
                    options = data[key]
                elif data[key] != True:
                    if key == 'crypt' and isinstance(data[key], basestring):
                        options = {'algorithm': data[key]}
                    else:
                        raise SyntaxError(
                            key+" validator accepts only dict or True")
                validators.append(vclass(**options))
        #: parse 'unique'
        _unique = data.get('unique', False)
        if _unique:
            _udbset = None
            if isinstance(_unique, dict):
                whr = _unique.get('where', None)
                if callable(whr):
                    _dbset = whr
            validators.append(
                notInDB(field.db, field.table, field.name, dbset=_udbset))
            table = field.db[field._tablename]
            table._unique_fields_validation_[field.name] = 1
        #: common options ('format', 'message')
        if 'format' in data:
            for validator in validators:
                children = [validator]
                if hasattr(validator, 'children'):
                    children += validator.children
                for child in children:
                    if hasattr(child, 'format'):
                        child.format = data['format']
                        break
        if 'message' in data:
            for validator in validators:
                validator.message = data['message']
        #: parse 'any'
        if 'any' in data:
            validators.append(Any(self(field, data['any'])))
        #: parse 'not'
        if 'not' in data:
            validators.append(Not(self(field, data['not'])))
        #: insert presence/empty validation if needed
        if presence:
            ref_table, multiple = self.parse_reference(field)
            if ref_table:
                if not _dbset:
                    validators.append(
                        inDB(field.db, ref_table, multiple=multiple))
            else:
                validators.insert(0, isntEmpty())
        if empty:
            validators.insert(0, isEmpty())
        #: parse 'allow'
        if 'allow' in data:
            if data['allow'] in ['empty', 'blank']:
                validators = [isEmptyOr(validators)]
            else:
                validators = [Allow(data['allow'], validators)]
        return validators



# -*- coding: utf-8 -*-
"""
    tests.utils
    ----------------

    Test weppy utils engine

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from weppy.datastructures import sdict
from weppy.utils import cachedprop, dict_to_sdict, is_valid_ip_address


class TestClass(object):

    @cachedprop
    def prop(self):
        return 'test_cachedprop'


def test_cachedprop():
    assert isinstance(TestClass.prop, cachedprop)
    obj = TestClass()
    assert obj.prop == 'test_cachedprop'


def test_is_valid_ip_address():
    result_localhost = is_valid_ip_address('127.0.0.1')
    assert result_localhost is True

    result_unknown = is_valid_ip_address('unknown')
    assert result_unknown is False

    result_ipv4_valid = is_valid_ip_address('::ffff:192.168.0.1')
    assert result_ipv4_valid is True

    result_ipv4_valid = is_valid_ip_address('192.168.256.1')
    assert result_ipv4_valid is False

    result_ipv6_valid = is_valid_ip_address('fd40:363d:ee85::')
    assert result_ipv6_valid is True

    result_ipv6_valid = is_valid_ip_address('fd40:363d:ee85::1::')
    assert result_ipv6_valid is False


def test_dict_to_sdict():
    result_sdict = dict_to_sdict({'test': 'dict'})
    assert isinstance(result_sdict, sdict)
    assert result_sdict.test == 'dict'

    result_number = dict_to_sdict(1)
    assert not isinstance(result_number, sdict)
    assert result_number == 1



# -*- coding: utf-8 -*-
"""
    weppy.tools.auth.models
    -----------------------

    Provides models for the authorization system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from datetime import datetime
from ..._compat import iterkeys
from ...dal import Model, Field, before_insert, rowmethod
from ...globals import current, request
from ...security import uuid


def _now():
    if hasattr(current, 'request'):
        return request.now
    return datetime.utcnow()


class TimestampedModel(Model):
    created_at = Field('datetime', default=_now, rw=False)
    updated_at = Field('datetime', default=_now, update=_now, rw=False)


class AuthModel(Model):
    auth = None

    form_registration_rw = {}
    form_profile_rw = {}

    def __super_method(self, name):
        return getattr(super(AuthModel, self), '_Model__' + name)

    def _define_(self):
        self.__super_method('define_validation')()
        self.__super_method('define_defaults')()
        self.__super_method('define_updates')()
        self.__super_method('define_representation')()
        self.__super_method('define_computations')()
        self.__super_method('define_callbacks')()
        self.__super_method('define_scopes')()
        self.__super_method('define_indexes')()
        self.__hide_all()
        self.__super_method('define_form_utils')
        self.__define_authform_utils()
        self.setup()

    #def __define_extra_fields(self):
    #    self.auth.settings.extra_fields['auth_user'] = self.fields

    def __hide_all(self):
        alwaysvisible = ['first_name', 'last_name', 'password', 'email']
        for field in self.table.fields:
            if field not in alwaysvisible:
                self.table[field].writable = self.table[field].readable = \
                    False

    def __base_visibility(self, form_type):
        exclude = []
        if form_type == 'profile_fields':
            exclude.append('password')
            exclude.append('email')
        return [
            field.name for field in self.table if
            field.writable and field.name not in exclude]

    def __define_authform_utils(self):
        settings_map = {
            'register_fields': 'form_registration_rw',
            'profile_fields': 'form_profile_rw'
        }
        for setting, attr in settings_map.items():
            rwdata = self.auth.settings[setting] or \
                self.__base_visibility(setting)
            if not isinstance(rwdata, dict):
                rwdata = {'writable': list(rwdata), 'readable': list(rwdata)}
            for field, value in getattr(self, attr).items():
                if isinstance(value, (tuple, list)):
                    readable, writable = value
                else:
                    readable = writable = value
                if readable:
                    rwdata['readable'].append(field)
                else:
                    if field in rwdata['readable']:
                        rwdata['readable'].remove(field)
                if writable:
                    rwdata['writable'].append(field)
                else:
                    if field in rwdata['writable']:
                        rwdata['writable'].remove(field)
            for key in iterkeys(rwdata):
                rwdata[key] = list(set(rwdata[key]))
            self.auth.settings[setting] = rwdata


class AuthUserBasic(AuthModel, TimestampedModel):
    tablename = "auth_users"
    format = '%(email)s (%(id)s)'
    #: injected by Auth
    #  has_many(
    #      {'memberships': 'AuthMembership'},
    #      {'authevents': 'AuthEvent'},
    #      {'authgroups': {'via': 'memberships'}},
    #      {'permissions': {'via': 'authgroups'}},
    #  )

    email = Field(length=255, unique=True)
    password = Field('password', length=512)
    registration_key = Field(length=512, rw=False, default='')
    reset_password_key = Field(length=512, rw=False, default='')
    registration_id = Field(length=512, rw=False, default='')

    form_labels = {
        'email': 'E-mail',
        'password': 'Password'
    }

    @before_insert
    def set_registration_key(self, fields):
        if self.auth.settings.registration_requires_verification and not \
                fields.get('registration_key'):
            fields['registration_key'] = uuid()

    @rowmethod('disable')
    def _set_disabled(self, row):
        return row.update_record(registration_key='disabled')

    @rowmethod('block')
    def _set_blocked(self, row):
        return row.update_record(registration_key='blocked')

    @rowmethod('allow')
    def _set_allowed(self, row):
        return row.update_record(registration_key='')


class AuthUser(AuthUserBasic):
    format = '%(first_name)s %(last_name)s (%(id)s)'

    first_name = Field(length=128, notnull=True)
    last_name = Field(length=128, notnull=True)

    form_labels = {
        'first_name': 'First name',
        'last_name': 'Last name',
    }


class AuthGroup(TimestampedModel):
    format = '%(role)s (%(id)s)'
    #: injected by Auth
    #  has_many(
    #      {'memberships': 'AuthMembership'},
    #      {'permissions': 'AuthPermission'},
    #      {'users': {'via': 'memberships'}}
    #  )

    role = Field(length=255, default='', unique=True)
    description = Field('text')

    form_labels = {
        'role': 'Role',
        'description': 'Description'
    }


class AuthMembership(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'user': 'AuthUser'}, {'authgroup': 'AuthGroup'})
    pass


class AuthPermission(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'authgroup': 'AuthGroup'})

    name = Field(length=512, default='default', notnull=True)
    table_name = Field(length=512)
    record_id = Field('int', default=0)

    validation = {
        'record_id': {'in': {'range': (0, 10**9)}}
    }

    form_labels = {
        'name': 'Name',
        'table_name': 'Object or table name',
        'record_id': 'Record ID'
    }


class AuthEvent(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'user': 'AuthUser'})

    client_ip = Field()
    origin = Field(length=512, notnull=True)
    description = Field('text', notnull=True)

    default_values = {
        'client_ip': lambda:
            request.client if hasattr(current, 'request') else 'unavailable',
        'origin': 'auth',
        'description': ''
    }

    #: labels injected by Auth
    form_labels = {
        'client_ip': 'Client IP',
        'origin': 'Origin',
        'description': 'Description'
    }


"""
class AuthUserSigned(AuthUser):
    is_active = Field('bool', default=True, rw=False)
    created_on = Field('datetime', default=lambda: datetime.utcnow(), rw=False)
    created_by = Field('reference auth_user', default=auth.user_id, rw=False)
    modified_on = Field('datetime', default=lambda: datetime.utcnow(),
                        update=lambda: datetime.utcnow(), rw=False),
    modified_by = Field('reference auth_user', default=auth.user_id,
                        update=auth.user_id, rw=False)
"""



# -*- coding: utf-8 -*-
"""
    weppy.tools.service
    -------------------

    Provides the services handler.

    :copyright: (c) 2014 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ..handlers import Handler
from ..globals import response
from ..serializers import json, xml


class ServiceHandler(Handler):
    def __init__(self, procedure):
        self.procedure = procedure

    def json(self, f, **kwargs):
        response.headers['Content-Type'] = 'application/json; charset=utf-8'
        data = f(**kwargs)
        return json(data)

    def xml(self, f, **kwargs):
        response.headers['Content-Type'] = 'text/xml'
        data = f(**kwargs)
        return xml(data)

    def wrap_call(self, func):
        def wrap(**kwargs):
            if hasattr(self, self.procedure):
                return self.__getattribute__(self.procedure)(func, **kwargs)
            else:
                raise RuntimeError(
                    'weppy cannot handle the service you requested: %s' %
                    self.procedure
                )
        return wrap



# -*- coding: utf-8 -*-
"""
    weppy.serializers
    -----------------

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import datetime
import decimal
import json as json_parser
import udatetime
from ._compat import PY2, integer_types
from .language.translator import TElement
from .tags import TAG, tag, htmlescape
from .datastructures import sdict


def _custom_json(o):
    if hasattr(o, 'custom_json') and callable(o.custom_json):
        return o.custom_json()
    if isinstance(o, datetime.datetime):
        return udatetime.to_string(o)
    elif isinstance(o, (datetime.date, datetime.time)):
        return o.isoformat()
    elif isinstance(o, integer_types):
        return int(o)
    elif isinstance(o, decimal.Decimal):
        return str(o)
    elif isinstance(o, TElement):
        return str(o)
    elif isinstance(o, TAG):
        return str(o)
    elif hasattr(o, 'as_list') and callable(o.as_list):
        return o.as_list()
    elif hasattr(o, 'as_dict') and callable(o.as_dict):
        return o.as_dict()
    else:
        raise TypeError(repr(o) + " is not JSON serializable")


def json(value, default=_custom_json):
    rep28 = r'\u2028'
    rep29 = r'\2029'
    if PY2:
        rep28 = rep28.decode('raw_unicode_escape')
        rep29 = rep29.decode('raw_unicode_escape')
    return json_parser.dumps(
        value, default=default
    ).replace(rep28, '\\u2028').replace(rep29, '\\u2029')


def _xml_rec(value, key, quote=True):
    if hasattr(value, 'custom_xml') and callable(value.custom_xml):
        return value.custom_xml()
    elif isinstance(value, (dict, sdict)):
        return tag[key](*[tag[k](_xml_rec(v, '', quote))
                          for k, v in value.items()])
    elif isinstance(value, list):
        return tag[key](*[tag[item](_xml_rec(item, '', quote))
                        for item in value])
    elif hasattr(value, 'as_list') and callable(value.as_list):
        return str(_xml_rec(value.as_list(), '', quote))
    elif hasattr(value, 'as_dict') and callable(value.as_dict):
        return str(_xml_rec(value.as_dict(), '', quote))
    else:
        return htmlescape(value, quote)


def xml(value, encoding='UTF-8', key='document', quote=True):
    return ('<?xml version="1.0" encoding="%s"?>' % encoding) + \
        str(_xml_rec(value, key, quote))



# -*- coding: utf-8 -*-
"""
    weppy.dal.base
    --------------

    Provides base pyDAL implementation for weppy.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import copy
import os
from pydal import DAL as _pyDAL, Field as _Field
from pydal._globals import THREAD_LOCAL
from pydal.objects import Table as _Table, Set as _Set, Rows as _Rows, \
    Expression, Row
from .._compat import copyreg, iterkeys, iteritems
from ..datastructures import sdict
from ..globals import current
from ..handlers import Handler
from ..security import uuid as _uuid
from ..serializers import _custom_json, xml
from ..utils import cachedprop
from ..validators import ValidateFromDict


class DALHandler(Handler):
    def __init__(self, db):
        self.db = db

    def on_start(self):
        self.db._adapter.reconnect()

    def on_success(self):
        self.db.commit()

    def on_failure(self):
        self.db.rollback()

    def on_end(self):
        self.db._adapter.close()


class Table(_Table):
    def __init__(self, *args, **kwargs):
        super(Table, self).__init__(*args, **kwargs)
        self._unique_fields_validation_ = {}

    def _create_references(self):
        self._referenced_by = []
        self._referenced_by_list = []
        self._references = []


class Set(_Set):
    def __init__(self, db, query, ignore_common_filters=None, model=None):
        super(Set, self).__init__(db, query, ignore_common_filters)
        self._model_ = model
        self._scopes_ = {}
        self._load_scopes_()

    def _load_scopes_(self):
        if self._model_ is None:
            tables = self.db._adapter.tables(self.query)
            if len(tables) == 1:
                self._model_ = self.db[tables[0]]._model_
        if self._model_:
            self._scopes_ = self._model_._scopes_

    def where(self, query, ignore_common_filters=False, model=None):
        if query is None:
            return self
        elif isinstance(query, Table):
            query = self.db._adapter.id_query(query)
        elif isinstance(query, str):
            query = Expression(self.db, query)
        elif isinstance(query, Field):
            query = query != None
        q = self.query & query if self.query else query
        return Set(
            self.db, q, ignore_common_filters=ignore_common_filters,
            model=model)

    def select(self, *fields, **options):
        pagination = options.get('paginate')
        if pagination:
            if isinstance(pagination, tuple):
                offset = pagination[0]
                limit = pagination[1]
            else:
                offset = pagination
                limit = 10
            options['limitby'] = ((offset - 1) * limit, offset * limit)
            del options['paginate']
        including = options.get('including')
        if including and self._model_ is not None:
            from .helpers import LeftJoinSet
            options['left'], jdata = self._parse_left_rjoins(including)
            del options['including']
            #: add fields to select
            fields = list(fields)
            if not fields:
                fields = [self._model_.table.ALL]
            for join in options['left']:
                fields.append(join.first.ALL)
            return LeftJoinSet._from_set(
                self, jdata).select(*fields, **options)
        return super(Set, self).select(*fields, **options)

    def validate_and_update(self, **update_fields):
        tablename = self.db._adapter.get_table(self.query)
        table = self.db[tablename]
        current._dbvalidation_record_id_ = None
        if table._unique_fields_validation_ and self.count() == 1:
            if any(
                table._unique_fields_validation_.get(fieldname)
                for fieldname in iterkeys(update_fields)
            ):
                current._dbvalidation_record_id_ = \
                    self.select(table.id).first().id
        response = Row()
        response.errors = Row()
        new_fields = copy.copy(update_fields)
        for key, value in iteritems(update_fields):
            value, error = self.db[tablename][key].validate(value)
            if error:
                response.errors[key] = '%s' % error
            else:
                new_fields[key] = value
        del current._dbvalidation_record_id_
        if response.errors:
            response.updated = None
        else:
            if not any(f(self, new_fields) for f in table._before_update):
                table._attempt_upload(new_fields)
                fields = table._listify(new_fields, update=True)
                if not fields:
                    raise SyntaxError("No fields to update")
                ret = self.db._adapter.update(tablename, self.query, fields)
                ret and [f(self, new_fields) for f in table._after_update]
            else:
                ret = 0
            response.updated = ret
        return response

    def join(self, *args):
        rv = self
        if self._model_ is not None:
            joins = []
            jtables = []
            for arg in args:
                condition, table, is_single_row = self._parse_rjoin(arg)
                joins.append(condition)
                jtables.append((arg, table._tablename, is_single_row))
            if joins:
                from .helpers import JoinSet
                q = joins[0]
                for join in joins[1:]:
                    q = q & join
                rv = rv.where(q)
                return JoinSet._from_set(rv, self._model_.tablename, jtables)
        return rv

    def _parse_rjoin(self, arg):
        from .helpers import RelationBuilder
        #: match has_many
        rel = self._model_._hasmany_ref_.get(arg)
        if rel:
            if isinstance(rel, dict) and rel.get('via'):
                r = RelationBuilder(rel, self._model_).via()
                return r[0], r[1]._table, False
            else:
                r = RelationBuilder(rel, self._model_)
                return r.many(), rel.table, False
        #: match belongs_to and refers_to
        rel = self._model_._belongs_ref_.get(arg)
        if rel:
            r = RelationBuilder((rel, arg), self._model_).belongs_query()
            return r, self._model_.db[rel], True
        #: match has_one
        rel = self._model_._hasone_ref_.get(arg)
        if rel:
            r = RelationBuilder(rel, self._model_)
            return r.many(), rel.table, True
        raise RuntimeError(
            'Unable to find %s relation of %s model' %
            (arg, self._model_.__name__))

    def _parse_left_rjoins(self, args):
        if not isinstance(args, (list, tuple)):
            args = [args]
        joins = []
        jdata = []
        for arg in args:
            condition, table, is_single_row = self._parse_rjoin(arg)
            joins.append(table.on(condition))
            jdata.append((arg, is_single_row))
        return joins, jdata

    def _jcolnames_from_rowstmps(self, tmps):
        colnames = []
        all_colnames = {}
        jcolnames = {}
        for colname in tmps:
            all_colnames[colname[0]] = colname[0]
            jcolnames[colname[0]] = jcolnames.get(colname[0], [])
            jcolnames[colname[0]].append(colname[1])
        for colname in iterkeys(all_colnames):
            if colname == self._stable_:
                colnames.append(colname)
                del jcolnames[colname]
        return colnames, jcolnames

    def __getattr__(self, name):
        scope = self._scopes_.get(name)
        if scope:
            from .helpers import ScopeWrap
            return ScopeWrap(self, self._model_, scope.f)
        raise AttributeError(name)


class Rows(_Rows):
    def __init__(self, db=None, records=[], colnames=[], compact=True,
                 rawrows=None):
        self.db = db
        self.records = records
        self.colnames = colnames
        self._rowkeys_ = list(self.records[0].keys()) if self.records else []
        self._getrow = self._getrow_compact_ if self.compact else self._getrow_

    @cachedprop
    def compact(self):
        if not self.records:
            return False
        return len(self._rowkeys_) == 1 and self._rowkeys_[0] != '_extra'

    @cachedprop
    def compact_tablename(self):
        if not self._rowkeys_:
            return None
        return self._rowkeys_[0]

    def _getrow_(self, i):
        return self.records[i]

    def _getrow_compact_(self, i):
        return self._getrow_(i)[self.compact_tablename]

    def __getitem__(self, i):
        return self._getrow(i)

    def column(self, column=None):
        colname = str(column) if column else self.colnames[0]
        return [r[colname] for r in self]

    def sorted(self, f, reverse=False):
        if self.compact:
            keyf = lambda r: f(r[self.compact_tablename])
        else:
            keyf = f
        return [r for r in sorted(self.records, key=keyf, reverse=reverse)]

    def sort(self, f, reverse=False):
        self.records = self.sorted(f, reverse)

    def append(self, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.append(row)

    def insert(self, position, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.insert(position, row)


class DAL(_pyDAL):
    serializers = {'json': _custom_json, 'xml': xml}
    logger = None
    uuid = lambda x: _uuid()

    record_operators = {}
    execution_handlers = []

    Rows = Rows

    @staticmethod
    def uri_from_config(config=None):
        if config is None or config.adapter is None:
            config = sdict(adapter="sqlite", host="dummy.db")
        if config.adapter == "<zombie>":
            return config.adapter
        if config.adapter == "sqlite" and config.host == "memory":
            return config.adapter + ":" + config.host
        uri = config.adapter + "://"
        if config.user:
            uri = uri + config.user + ":" + config.password + "@"
        uri = uri + config.host
        if config.database:
            uri += "/" + config.database
        return uri

    def __new__(cls, app, *args, **kwargs):
        config = kwargs.get('config', sdict()) or app.config.db
        uri = config.uri or DAL.uri_from_config(config)
        return super(DAL, cls).__new__(cls, uri, *args, **kwargs)

    def __init__(self, app, config=sdict(), pool_size=None, folder=None,
                 **kwargs):
        self.logger = app.log
        config = config or app.config.db
        if not config.uri:
            config.uri = self.uri_from_config(config)
        self.config = config
        #: load config data
        kwargs['check_reserved'] = self.config.check_reserved or \
            kwargs.get('check_reserved', None)
        kwargs['migrate'] = self.config.auto_migrate or \
            kwargs.get('auto_migrate', True)
        kwargs['driver_args'] = self.config.driver_args or \
            kwargs.get('driver_args', None)
        kwargs['adapter_args'] = self.config.adapter_args or \
            kwargs.get('adapter_args', None)
        if kwargs.get('auto_migrate') is not None:
            del kwargs['auto_migrate']
        #: set directory
        folder = folder or 'databases'
        folder = os.path.join(app.root_path, folder)
        if not os.path.exists(folder):
            os.mkdir(folder)
        #: set pool_size
        pool_size = self.config.pool_size or pool_size or 0
        #: add timings storage if requested
        if config.store_execution_timings:
            from .helpers import TimingHandler
            self.execution_handlers.append(TimingHandler)
        #: finally setup pyDAL instance
        super(DAL, self).__init__(self.config.uri, pool_size, folder, **kwargs)
        self._adapter._add_operators_to_parsed_row = \
            lambda *args, **kwargs: None
        self._adapter._add_reference_sets_to_parsed_row = \
            lambda *args, **kwargs: None

    @property
    def handler(self):
        return DALHandler(self)

    @property
    def execution_timings(self):
        return getattr(THREAD_LOCAL, '_weppydal_timings_', [])

    def define_models(self, *models):
        if len(models) == 1 and isinstance(models[0], (list, tuple)):
            models = models[0]
        for model in models:
            if not hasattr(self, model.__name__):
                # store db instance inside model
                model.db = self
                # init model
                obj = model()
                obj._define_props_()
                obj._define_relations_()
                obj._define_virtuals_()
                # define table and store in model
                #model.fields = obj.fields
                args = dict(
                    migrate=obj.migrate,
                    format=obj.format,
                    table_class=Table
                )
                model.table = self.define_table(
                    obj.tablename, *obj.fields, **args
                )
                model.table._model_ = obj
                model.id = model.table.id
                # load user's definitions
                obj._define_()
                # set reference in db for model name
                self.__setattr__(model.__name__, obj.table)

    def where(self, query=None, ignore_common_filters=None, model=None):
        q = None
        if isinstance(query, Table):
            q = self._adapter.id_query(query)
        elif isinstance(query, Field):
            q = (query != None)
        elif isinstance(query, dict):
            icf = query.get("ignore_common_filters")
            if icf:
                ignore_common_filters = icf
        if q is None and query is not None:
            if hasattr(query, '_belongs_ref_'):
                q = self._adapter.id_query(query.table)
            else:
                q = query
        return Set(
            self, q, ignore_common_filters=ignore_common_filters, model=model)


def _DAL_unpickler(db_uid):
    fake_app_obj = sdict(config=sdict(db=sdict()))
    fake_app_obj.config.db.adapter = '<zombie>'
    return DAL(fake_app_obj, db_uid=db_uid)


def _DAL_pickler(db):
    return _DAL_unpickler, (db._db_uid,)

copyreg.pickle(DAL, _DAL_pickler, _DAL_unpickler)


class Field(_Field):
    _weppy_types = {
        'integer': 'int', 'double': 'float', 'boolean': 'bool',
        'list:integer': 'list:int'
    }
    _pydal_types = {
        'int': 'integer', 'bool': 'boolean', 'list:int': 'list:integer'
    }
    _weppy_delete = {
        'cascade': 'CASCADE', 'nullify': 'SET NULL', 'nothing': 'NO ACTION'
    }
    _inst_count_ = 0
    _obj_created_ = False

    def __init__(self, type='string', *args, **kwargs):
        self.modelname = None
        self._auto_validation = True
        #: convert type
        self._type = self._weppy_types.get(type, type)
        #: convert 'rw' -> 'readable', 'writeable'
        if 'rw' in kwargs:
            if isinstance(kwargs['rw'], (tuple, list)):
                read, write = kwargs['rw']
            else:
                read = write = kwargs['rw']
            kwargs['readable'] = read
            kwargs['writable'] = write
            del kwargs['rw']
        #: convert 'info' -> 'comment'
        _info = kwargs.get('info')
        if _info:
            kwargs['comment'] = _info
            del kwargs['info']
        #: convert ondelete parameter
        _ondelete = kwargs.get('ondelete')
        if _ondelete:
            if _ondelete not in list(self._weppy_delete):
                raise SyntaxError(
                    'Field ondelete should be set on %s, %s or %s' %
                    list(self._weppy_delete)
                )
            kwargs['ondelete'] = self._weppy_delete[_ondelete]
        #: process 'refers_to' fields
        self._isrefers = kwargs.get('_isrefers')
        if self._isrefers:
            del kwargs['_isrefers']
        #: get auto validation preferences
        if 'auto_validation' in kwargs:
            self._auto_validation = kwargs['auto_validation']
            del kwargs['auto_validation']
        #: intercept validation (will be processed by `_make_field`)
        self._requires = {}
        self._custom_requires = []
        if 'validation' in kwargs:
            if isinstance(kwargs['validation'], dict):
                self._requires = kwargs['validation']
            else:
                self._custom_requires = kwargs['validation']
                if not isinstance(self._custom_requires, list):
                    self._custom_requires = [self._custom_requires]
            del kwargs['validation']
        self._validation = {}
        self._vparser = ValidateFromDict()
        #: store args and kwargs for `_make_field`
        self._args = args
        self._kwargs = kwargs
        #: increase creation counter (used to keep order of fields)
        self._inst_count_ = Field._inst_count_
        Field._inst_count_ += 1

    def _default_validation(self):
        rv = {}
        auto_types = [
            'int', 'float', 'decimal', 'date', 'time', 'datetime', 'json'
        ]
        if self._type in auto_types:
            rv['is'] = self._type
        if self._type == 'bigint':
            rv['is'] = 'int'
        if self._type == 'bool':
            rv['in'] = (False, True)
        if self._type in ['string', 'text', 'password']:
            rv['len'] = {'lte': self.length}
        if self._type == 'password':
            rv['len']['gte'] = 6
            rv['crypt'] = True
        if self._type == 'list:int':
            rv['is'] = 'list:int'
        if self.notnull or self._type.startswith('reference') or \
                self._type.startswith('list:reference'):
            rv['presence'] = True
        if not self.notnull and self._isrefers is True:
            rv['allow'] = 'empty'
        if self.unique:
            rv['unique'] = True
        return rv

    def _parse_validation(self):
        for key in list(self._requires):
            self._validation[key] = self._requires[key]
        self.requires = self._vparser(self, self._validation) + \
            self._custom_requires

    #: `_make_field` will be called by `Model` class or `Form` class
    #  it will make weppy's Field class compatible with the pyDAL's one
    def _make_field(self, name, model=None):
        if self._obj_created_:
            return self
        if model is not None:
            self.modelname = model.__class__.__name__
        #: convert field type to pyDAL ones if needed
        ftype = self._pydal_types.get(self._type, self._type)
        #: create pyDAL's Field instance
        super(Field, self).__init__(name, ftype, *self._args, **self._kwargs)
        #: add automatic validation (if requested)
        if self._auto_validation:
            auto = True
            if self.modelname:
                auto = model.auto_validation
            if auto:
                self._validation = self._default_validation()
        #: validators
        if not self.modelname:
            self._parse_validation()
        self._obj_created_ = True
        return self

    def __str__(self):
        if self._obj_created_:
            return super(Field, self).__str__()
        return object.__str__(self)

    def __repr__(self):
        if self.modelname and hasattr(self, 'name'):
            return "<%s.%s (%s) field>" % (self.modelname, self.name,
                                           self._type)
        return super(Field, self).__repr__()



# -*- coding: utf-8 -*-
"""
    tests.cache
    ----------------

    Test weppy cache module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""


from weppy.cache import BaseCache, RamCache, DiskCache, Cache


def test_basecache():
    base_cache = BaseCache()
    assert base_cache._default_expire == 300

    assert base_cache('key') is None
    assert base_cache._get('key') is None
    assert base_cache._set('key', 'value', 300) is None
    assert base_cache.clear() is None


def test_ramcache():
    ram_cache = RamCache()
    assert ram_cache._prefix == ''
    assert ram_cache._threshold == 500

    ram_cache('test', lambda: 2)
    assert ram_cache._get('test') == 2

    ram_cache.clear()
    assert ram_cache._get('test') is None


def test_diskcache():
    disk_cache = DiskCache()
    assert disk_cache._threshold == 500

    disk_cache('test', lambda: 2)
    assert disk_cache._get('test') == 2

    disk_cache.clear()
    assert disk_cache._get('test') is None


def test_cache():
    default_cache = Cache()
    assert default_cache.default_handler == 'ram'

    default_cache('test', lambda: 2)
    assert default_cache('test', lambda: 3) == 2

    disk_cache = DiskCache()
    ram_cache = RamCache()
    cache = Cache(default='disc', ram=ram_cache, disc=disk_cache)
    assert cache.default_handler == 'disc'
    assert cache.disc == disk_cache
    assert cache.ram == ram_cache



# -*- coding: utf-8 -*-
"""
    tests.templates
    ---------------

    Test weppy templating module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import pytest
from weppy import App
from weppy.globals import current


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    return app


def test_define(app):
    templater = app.templater
    assert templater._render(source='{{=1}}', filename='test1') == '1'
    assert templater._render(
        source='{{=a}}', filename='test2',
        context={'a': 'nuvolosità variabile'}
    ) == 'nuvolosit&#224; variabile'
    assert templater._render(
        source='{{=a}}', path='templates', filename='test3',
        context={'a': u'nuvolosità variabile'}
    ) == 'nuvolosit&#224; variabile'
    assert templater._render(
        source='{{=a}}', filename='test4',
        context={'a': [i for i in range(0, 5)]}
    ) == "[0, 1, 2, 3, 4]"


# def test_superblock(app):
#     pass


def test_helpers(app):
    templater = app.templater
    r = templater._render(source="{{include_helpers}}", filename='testh')
    assert r == '<script type="text/javascript" ' + \
        'src="/__weppy__/jquery.min.js"></script>\n' + \
        '<script type="text/javascript" ' + \
        'src="/__weppy__/helpers.js"></script>'


def test_meta(app):
    current.initialize({
        'PATH_INFO': '/', 'wpp.appnow': '', 'wpp.now.utc': '',
        'wpp.now.local': '', 'wpp.application': ''})
    current.response.meta.foo = "bar"
    current.response.meta_prop.foo = "bar"
    templater = app.templater
    r = templater._render(
        source="{{include_meta}}", filename='mtest',
        context={'current': current})
    assert r == '<meta name="foo" content="bar" />\n' + \
        '<meta property="foo" content="bar" />\n'


def test_static(app):
    templater = app.templater
    s = "{{include_static 'foo.js'}}\n{{include_static 'bar.css'}}"
    r = templater._render(source=s, filename="stest")
    assert r == '<script type="text/javascript" src="/static/foo.js">' + \
        '</script>\n<link rel="stylesheet" href="/static/bar.css" ' + \
        'type="text/css" />'


def test_pycode(app):
    templater = app.templater
    #: test if block
    s = "{{if a == 1:}}foo{{elif a == 2:}}bar{{else:}}foobar{{pass}}"
    r = templater._render(source=s, filename="ptest1", context={'a': 1})
    assert r == "foo"
    r = templater._render(source=s, filename="ptest1", context={'a': 2})
    assert r == "bar"
    r = templater._render(source=s, filename="ptest1", context={'a': 25})
    assert r == "foobar"
    #: test for block
    s = "{{for i in range(0, 5):}}{{=i}}\n{{pass}}"
    r = templater._render(source=s, filename="ptest2")
    assert r == "0\n1\n2\n3\n4\n"


rendered_value = """
<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>
        <meta name="foo" content="bar" />
        <meta property="foo" content="bar" />
        <script type="text/javascript" src="/__weppy__/jquery.min.js"></script>
        <script type="text/javascript" src="/__weppy__/helpers.js"></script>
        <link rel="stylesheet" href="/static/style.css" type="text/css" />
    </head>
    <body>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">nuvolosit&#224; variabile</a>
            </div>
            <ul class="posts">
                <li>
                    <h2>foo</h2>
                    <hr />
                </li>

                <li>
                    <h2>bar</h2>
                    <hr />
                </li>
            </ul>
        </div>
    </body>
</html>
"""


def test_render(app):
    current._language = 'it'
    r = app.templater.render(
        app.template_path, 'test.html', {
            'current': current, 'posts': [{'title': 'foo'}, {'title': 'bar'}]
        }
    )
    assert "".join([l.strip() for l in r.splitlines()]) == \
        "".join([l.strip() for l in rendered_value.splitlines()])


# def test_cache(app):
#     pass
